#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Optional, Tuple

import numpy as np
import pinocchio as pin
import upkie_description
from gymnasium import spaces

from upkie.utils.clamp import clamp_and_warn
from upkie.utils.exceptions import ActionError, ModelError
from upkie.utils.pinocchio import (
    box_position_limits,
    box_torque_limits,
    box_velocity_limits,
)
from upkie.utils.robot_state import RobotState

from .upkie_base_env import UpkieBaseEnv


class UpkieServos(UpkieBaseEnv):

    """!
    Upkie with full observation and joint position-velocity-torque actions.

    ### Action space

    The action space is a dictionary with one key for each servo on Upkie:

    - ``left_hip``: Left hip joint (qdd100)
    - ``left_hip``: Left knee joint (qdd100)
    - ``left_hip``: Left wheel joint (mj5208)
    - ``right_hip``: Right hip joint (qdd100)
    - ``right_hip``: Right knee joint (qdd100)
    - ``right_hip``: Right wheel joint (mj5208)

    The value for each dictionary is itself a dictionary for each

    - ``position``: Joint angle in [rad] (NaN to disable) (required).
    - ``velocity``: Joint velocity in [rad] / [s] (required).
    - ``feedforward_torque``: Joint torque in [N] * [m].
    - ``kp_scale``: Scaling factor applied to the position feedback gain,
        between zero and one.
    - ``kd_scale``: Scaling factor applied to the velocity feedback gain,
        between zero and one.
    - ``maximum_torque``: Maximum joint torque (feedforward + feedback) during
        the whole actuation step, in [N] * [m].

    ### Observation space

    Obversations from this environment are full dictionary reported by the
    spine. See @ref observations.

    ### Attributes

    The environment has the following attributes:

    - ``robot``: Pinocchio robot wrapper.
    - ``version``: Environment version number.
    """

    ACTION_KEYS: Tuple[str, str, str, str, str, str] = (
        "position",
        "velocity",
        "feedforward_torque",
        "kp_scale",
        "kd_scale",
        "maximum_torque",
    )

    JOINT_NAMES: Tuple[str, str, str, str, str, str] = (
        "left_hip",
        "left_knee",
        "left_wheel",
        "right_hip",
        "right_knee",
        "right_wheel",
    )

    robot: pin.RobotWrapper
    version: int = 3

    def __init__(
        self,
        fall_pitch: float = 1.0,
        frequency: float = 200.0,
        init_state: Optional[RobotState] = None,
        regulate_frequency: bool = True,
        shm_name: str = "/vulp",
        spine_config: Optional[dict] = None,
    ):
        """!
        Initialize environment.

        @param fall_pitch Fall pitch angle, in radians.
        @param frequency Regulated frequency of the control loop, in Hz.
        @param init_state Initial state of the robot, only used in simulation.
        @param regulate_frequency Enables loop frequency regulation.
        @param shm_name Name of shared-memory file.
        @param spine_config Additional spine configuration overriding the
            defaults from ``//config:spine.yaml``. The combined configuration
            dictionary is sent to the spine at every :func:`reset`.
        """
        super().__init__(
            fall_pitch=fall_pitch,
            frequency=frequency,
            init_state=init_state,
            regulate_frequency=regulate_frequency,
            shm_name=shm_name,
            spine_config=spine_config,
        )

        robot = upkie_description.load_in_pinocchio(root_joint=None)
        model = robot.model
        q_min, q_max = box_position_limits(model)
        v_max = box_velocity_limits(model)
        tau_max = box_torque_limits(model)
        joint_names = list(model.names)[1:]
        if set(joint_names) != set(self.JOINT_NAMES):
            raise ModelError(
                "Upkie joints don't match:"
                f"expected {self.JOINT_NAMES} "
                f"got {joint_names}"
            )

        action_space = {}
        default_action = {}
        max_action = {}
        min_action = {}
        servo_space = {}
        for name in joint_names:
            joint = model.joints[model.getJointId(name)]
            action_space[name] = spaces.Dict(
                {
                    "position": spaces.Box(
                        low=q_min[joint.idx_q],
                        high=q_max[joint.idx_q],
                        shape=(1,),
                        dtype=float,
                    ),
                    "velocity": spaces.Box(
                        low=-v_max[joint.idx_v],
                        high=+v_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                    "feedforward_torque": spaces.Box(
                        low=-tau_max[joint.idx_v],
                        high=+tau_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                    "kp_scale": spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "kd_scale": spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "maximum_torque": spaces.Box(
                        low=0.0,
                        high=tau_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            servo_space[name] = spaces.Dict(
                {
                    "position": spaces.Box(
                        low=q_min[joint.idx_q],
                        high=q_max[joint.idx_q],
                        shape=(1,),
                        dtype=float,
                    ),
                    "velocity": spaces.Box(
                        low=-v_max[joint.idx_v],
                        high=+v_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                    "torque": spaces.Box(
                        low=-tau_max[joint.idx_v],
                        high=+tau_max[joint.idx_v],
                        shape=(1,),
                        dtype=float,
                    ),
                    "temperature": spaces.Box(
                        low=0.0,
                        high=100.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "voltage": spaces.Box(
                        low=10.0,  # moteus min 10 V
                        high=44.0,  # moteus max 44 V
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            default_action[name] = {
                # "position": np.nan,  # no default, needs to be explicit
                "velocity": 0.0,
                "feedforward_torque": 0.0,
                "kp_scale": 1.0,
                "kd_scale": 1.0,
                "maximum_torque": tau_max[joint.idx_v],
            }
            max_action[name] = {
                "position": q_max[joint.idx_q],
                "velocity": v_max[joint.idx_v],
                "feedforward_torque": tau_max[joint.idx_v],
                "kp_scale": 1.0,
                "kd_scale": 1.0,
                "maximum_torque": tau_max[joint.idx_v],
            }
            min_action[name] = {
                "position": q_min[joint.idx_q],
                "velocity": -v_max[joint.idx_v],
                "feedforward_torque": -tau_max[joint.idx_v],
                "kp_scale": 0.0,
                "kd_scale": 0.0,
                "maximum_torque": 0.0,
            }

        # gymnasium.Env: action_space
        self.action_space = spaces.Dict(action_space)

        # gymnasium.Env: observation_space
        self.observation_space = spaces.Dict(
            {
                "imu": spaces.Dict(
                    {
                        "angular_velocity": spaces.Box(
                            low=-np.inf,
                            high=np.inf,
                            shape=(3,),
                            dtype=float,
                        ),
                        "linear_acceleration": spaces.Box(
                            low=-np.inf,
                            high=np.inf,
                            shape=(3,),
                            dtype=float,
                        ),
                        "orientation": spaces.Box(
                            low=-1.0,
                            high=1.0,
                            shape=(4,),
                            dtype=float,
                        ),
                    }
                ),
                "servo": spaces.Dict(servo_space),
                "wheel_odometry": spaces.Dict(
                    {
                        "position": spaces.Box(
                            low=-np.inf,
                            high=np.inf,
                            shape=(1,),
                            dtype=float,
                        ),
                        "velocity": spaces.Box(
                            low=-np.inf,
                            high=np.inf,
                            shape=(1,),
                            dtype=float,
                        ),
                    }
                ),
            }
        )

        # Class members
        self.__default_action = default_action
        self.__max_action = max_action
        self.__min_action = min_action
        self.robot = robot

    def get_default_action(self) -> dict:
        """!
        Get a default action dictionary.

        @returns Default action dictionary.
        """
        return self.__default_action.copy()

    def get_env_observation(self, spine_observation: dict):
        """!
        Extract environment observation from spine observation dictionary.

        @param spine_observation Full observation dictionary from the spine.
        @returns Environment observation.
        """
        # If creating a new object turns out to be too slow we can switch to
        # updating in-place.
        return {
            "imu": {
                "angular_velocity": np.array(
                    spine_observation["imu"]["angular_velocity"],
                ),
                "linear_acceleration": np.array(
                    spine_observation["imu"]["linear_acceleration"]
                ),
                "orientation": np.array(
                    spine_observation["imu"]["orientation"]
                ),
            },
            "servo": {
                joint: {
                    key: spine_observation["servo"][joint][key]
                    for key in self.observation_space["servo"][joint]
                }
                for joint in self.JOINT_NAMES
            },
            "wheel_odometry": {
                "position": spine_observation["wheel_odometry"]["position"],
                "velocity": spine_observation["wheel_odometry"]["velocity"],
            },
        }

    def get_spine_action(self, env_action: dict) -> dict:
        """!
        Convert environment action to a spine action dictionary.

        @param env_action Environment action.
        @returns Spine action dictionary.
        """
        spine_action = {"servo": {}}
        for joint in self.JOINT_NAMES:
            servo_action = {}
            for key in self.ACTION_KEYS:
                try:
                    action = (
                        env_action[joint][key]
                        if key in env_action[joint]
                        else self.__default_action[joint][key]
                    )
                except KeyError as key_error:
                    raise ActionError(
                        f'Missing key "{key}" required for joint "{joint}"'
                    ) from key_error
                servo_action[key] = clamp_and_warn(
                    action,
                    self.__min_action[joint][key],
                    self.__max_action[joint][key],
                    label=f"{joint}: {key}",
                )
            spine_action["servo"][joint] = servo_action
        return spine_action

    def get_reward(self, observation: dict, action: dict) -> float:
        """!
        Get reward from observation and action.

        @param observation Environment observation.
        @param action Environment action.
        @returns Reward.
        """
        return 1.0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria
#
# This file incorporates work covered by the following copyright and
# permission notice:
#
#     setup-system.py from github.com:mjbots/quad
#     Copyright 2018-2020 Josh Pieper
#     SPDX-License-Identifier: Apache-2.0


"""Set up a Raspberry Pi 4b to operate a Wi-Fi access point.

The base operating system should already be installed. This script is intended
to be run as root, like: ``sudo ./setup-wifi-ap.py``.
"""

import argparse
import datetime
import os
import pathlib
import shutil
import subprocess
import time

ORIG_SUFFIX = time.strftime(".orig-%Y%m%d-%H%M%S")

# Log each setup step, in case
# ============================

logging_depth = 0


def log_message(message: str, indent: int = 0) -> None:
    """Log a single message.

    Args:
        message: Message to log.
        indent: Indentation level.
    """
    logging_indent = " | " * (logging_depth + indent)
    now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    print(f"{now}: {logging_indent}{message}")


def log_method(func):
    """Decorator to log function calls."""

    def wrapped_function(*args, **kwargs):
        global logging_depth
        try:
            logging_depth += 1
            args_repr = (
                f'"{args[0]}"' + (", ..." if len(args) > 1 else "")
                if args
                else ""
            )
            log_message(f"{func.__name__}({args_repr})", indent=-1)
            result = func(*args, **kwargs)
            return result
        finally:
            logging_depth -= 1

    return wrapped_function


# Command-line arguments
# ======================


def parse_command_line_arguments() -> argparse.Namespace:
    """Parse command-line arguments.

    Returns
    -------
        Namespace resulting from parsing command-line arguments.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "ssid",
        help="SSID of the robot's Wi-Fi network",
    )
    parser.add_argument(
        "wpa_passphrase",
        help="Password of the robot's Wi-Fi network",
    )
    parser.add_argument(
        "country_code",
        help="Two-letter country code for Wi-Fi compliance (e.g. FR or UK)",
    )
    parser.add_argument(
        "wlan_prefix",
        help="IPv4 address prefix (first three bytes) "
        "for the robot's Wi-Fi network interface (for instance 192.168.0)",
    )
    parser.add_argument(
        "eth_prefix",
        help="IPv4 address prefix (first three bytes) "
        "for the robot's wired network interface (for instance 192.168.1)",
    )
    return parser.parse_args()


# Utility functions
# =================


@log_method
def run(*args, **kwargs):
    """Run a subprocess."""
    subprocess.check_call(*args, shell=True, **kwargs)


@log_method
def ensure_present(filename, line):
    """Ensure a given line is present in a named file, and add it if not.

    Args:
        filename: Path to file.
        line: Line that should be present in the file's content.
    """
    current_content = [
        x.strip() for x in open(filename, encoding="utf-8").readlines()
    ]
    if line.strip() in current_content:
        return
    shutil.copy(filename, filename + ORIG_SUFFIX)
    log_message(f"Adding: {line}")
    with open(filename, "a", encoding="utf-8") as f:
        f.write(line + "\n")


@log_method
def ensure_contents(filename, contents):
    """Ensure the given file has exactly the given contents.

    Args:
        filename: Path to file.
        contents: File contents.
    """
    pathlib.Path(filename).parent.mkdir(parents=True, exist_ok=True)

    if os.path.exists(filename):
        existing = open(filename, encoding="utf-8").read()
        if existing == contents:
            return
        shutil.copy(filename, filename + ORIG_SUFFIX)

    log_message("Updating file content")

    with open(filename, "w", encoding="utf-8") as f:
        f.write(contents)


@log_method
def set_config_var(name, value):
    """Set the given variable in /boot/config.txt.

    Args:
        name: Variable name.
        value: Value to set.
    """
    contents = open("/boot/config.txt", encoding="utf-8").readlines()

    new_value = "{}={}".format(name, value)

    maybe_value = [x for x in contents if x.startswith("{}=".format(name))]
    if len(maybe_value) == 1 and maybe_value[0].strip() == new_value:
        return

    new_contents = [
        x for x in contents if not x.startswith("{}=".format(name))
    ] + [new_value + "\n"]

    shutil.copy("/boot/config.txt", "/boot/config.txt" + ORIG_SUFFIX)

    log_message("Updating {name} to {value} in /boot/config.txt")

    open("/boot/config.txt", "w", encoding="utf-8").write(
        "".join([x for x in new_contents])
    )


# Setup steps
# ===========


@log_method
def install_packages():
    """Install packages for the access point."""
    run("apt-get install --yes hostapd dnsmasq")


@log_method
def configure_interfaces(wlan_prefix, eth_prefix):
    """Configure network interfaces.

    Args:
        wlan_prefix: IP prefix (XXX.YYY.ZZZ) for the wireless network.
        eth_prefix: IP prefix (XXX.YYY.ZZZ) for the wired network.
    """
    ensure_contents(
        "/etc/network/interfaces",
        """# interfaces(5) file used by ifup(8) and ifdown(8)

# Please note that this file is written to be used with dhcpcd
# For static IP, consult /etc/dhcpcd.conf and 'man dhcpcd.conf'

# Include files from /etc/network/interfaces.d:
source-directory /etc/network/interfaces.d
""",
    )
    ensure_contents(
        "/etc/dhcpcd.conf",
        f"""# Configuration for dhcpcd.
# See dhcpcd.conf(5) for details.

# Allow users of this group to interact with dhcpcd via the control socket.
#controlgroup wheel

# Inform the DHCP server of our hostname for DDNS.
hostname

# Use the hardware address of the interface for the Client ID.
clientid
# or
# Use the same DUID + IAID as set in DHCPv6 for DHCPv4 ClientID as per RFC4361.
# Some non-RFC compliant DHCP servers do not reply with this set.
# In this case, comment out duid and enable clientid above.
#duid

# Persist interface configuration when dhcpcd exits.
persistent

# Rapid commit support.
# Safe to enable by default because it requires the equivalent option set
# on the server to actually work.
option rapid_commit

# A list of options to request from the DHCP server.
option domain_name_servers, domain_name, domain_search, host_name
option classless_static_routes
# Respect the network MTU. This is applied to DHCP routes.
option interface_mtu

# Most distributions have NTP support.
#option ntp_servers

# A ServerID is required by RFC2131.
require dhcp_server_identifier

# Generate SLAAC address using the Hardware Address of the interface
#slaac hwaddr
# OR generate Stable Private IPv6 Addresses based from the DUID
slaac private

# Static IP configuration for debugging
interface eth0
static ip_address={eth_prefix}.42/24
static routers={eth_prefix}.1

# Wireless access point configuration for the robot
interface wlan0
nohook wpa_supplicant
static ip_address={wlan_prefix}.42/24
static routers={wlan_prefix}.1
""",
    )


@log_method
def configure_hostapd(ssid, wpa_passphrase, country_code):
    """Configure the Wi-Fi access point.

    Args:
        ssid: Name of the network (SSID).
        wpa_passphrase: Passphrase.
        country_code: Two-letter country code.
    """
    ensure_contents(
        "/etc/hostapd/hostapd.conf",
        f"""country_code={country_code}

interface=wlan0
driver=nl80211
ssid={ssid}
hw_mode=a
channel=36
ieee80211n=1
require_ht=1
ieee80211ac=1
require_vht=1
ieee80211d=1
ieee80211h=0

ht_capab=[HT40+][SHORT-GI-20][DSSS_CK-40][MAX-AMSDU-3839]
vht_capab=[MAX-MDPU-3895][SHORT-GI-80][SU-BEAMFORMEE]

vht_oper_chwidth=1
vht_oper_centr_freq_seg0_idx=42

wmm_enabled=0
macaddr_acl=0
auth_algs=1
ignore_broadcast_ssid=0

wpa=2
wpa_passphrase={wpa_passphrase}
wpa_key_mgmt=WPA-PSK
wpa_pairwise=TKIP
rsn_pairwise=CCMP
""",
    )
    ensure_present(
        "/etc/default/hostapd", 'DAEMON_CONF="/etc/hostapd/hostapd.conf"'
    )
    run("systemctl unmask hostapd")
    run("systemctl enable hostapd")
    run("systemctl start hostapd")
    run("systemctl daemon-reload")


@log_method
def configure_dnsmasq(wlan_prefix):
    """Configure dnsmasq.

    Args:
        wlan_prefix: IP prefix (XXX.YYY.ZZZ) for the wireless network.
    """
    ensure_contents(
        "/etc/dnsmasq.conf",
        f"""interface=wlan0
listen-address=::1,127.0.0.1,{wlan_prefix}.42
domain-needed
bogus-priv
dhcp-range={wlan_prefix}.100,{wlan_prefix}.110,255.255.255.0,24h
""",
    )


@log_method
def configure_access_point(
    ssid,
    wpa_passphrase,
    country_code,
    wlan_prefix,
    eth_prefix,
):
    """Configure the raspi as an access point.

    Args:
        ssid: SSID of the Wi-Fi network.
        wpa_passphrase: Wi-Fi password.
        country_code: Two-letter country code for compliance.
        wlan_prefix: IPv4 address prefix (three first bytes) for the robot's
            Wi-Fi network, e.g. "192.168.0".
        eth_prefix: IPv4 address prefix (three first bytes) for the robot's
            wired network, e.g. "192.168.1".

    Note:
        The robot's IP suffix on both eth0 and wlan0 interfaces is 42. IP
        address suffixes on the wireless network are assigned from 100 to 110.
    """
    run("rfkill unblock all")
    configure_interfaces(wlan_prefix, eth_prefix)
    configure_hostapd(ssid, wpa_passphrase, country_code)
    configure_dnsmasq(wlan_prefix)


if __name__ == "__main__":
    if os.getuid() != 0:
        raise RuntimeError("must be run as root")
    args = parse_command_line_arguments()
    install_packages()
    configure_access_point(
        args.ssid,
        args.wpa_passphrase,
        args.country_code,
        args.wlan_prefix,
        args.eth_prefix,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Tiny example: balancing using a proportional wheel controller."""

import gymnasium as gym
import upkie.envs

upkie.envs.register()

if __name__ == "__main__":
    with gym.make("UpkieGroundVelocity-v3", frequency=200.0) as env:
        observation, _ = env.reset()  # connects to the spine
        action = 0.0 * env.action_space.sample()
        for step in range(1_000_000):
            pitch = observation[0]
            action[0] = 10.0 * pitch  # 1D action: [ground_velocity]
            observation, reward, terminated, truncated, _ = env.step(action)
            if terminated or truncated:
                observation, _ = env.reset()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test upkie.envs submodule."""

import unittest

import gymnasium as gym
import posix_ipc

from upkie.envs import register


class TestUpkieEnvs(unittest.TestCase):
    def test_register(self, env_name: str = "UpkieServos"):
        with self.assertRaises(gym.error.NameNotFound):
            # runs in the same test so that we make sure this is executed
            # before the call to ``register()``
            gym.make(env_name)
        shm_name = "/foobar"
        shared_memory = posix_ipc.SharedMemory(
            shm_name, posix_ipc.O_RDWR | posix_ipc.O_CREAT, size=42
        )
        register()
        gym.make(env_name, shm_name=shm_name)
        shared_memory.close_fd()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test Pinocchio utility functions."""

import unittest

import numpy as np
import pinocchio as pin

from upkie.utils.pinocchio import (
    box_position_limits,
    box_torque_limits,
    box_velocity_limits,
)


class TestPinocchio(unittest.TestCase):
    def test_box_limits(self):
        model = pin.Model()
        model.addJoint(
            0,
            pin.JointModelRevoluteUnboundedUnaligned(),
            pin.SE3.Identity(),
            "J1",
        )
        model.addJoint(
            1,
            pin.JointModelRevoluteUnaligned(),
            pin.SE3.Identity(),
            "revolute",
            max_effort=np.array([1.0]),
            max_velocity=np.array([42.0]),
            min_config=np.array([-12.0]),
            max_config=np.array([12.0]),
        )
        model.addJoint(
            2,
            pin.JointModelRevoluteUnaligned(),
            pin.SE3.Identity(),
            "revolute",
            max_effort=np.array([0.0]),
            max_velocity=np.array([0.0]),
            min_config=np.array([0.0]),
            max_config=np.array([0.0]),
        )
        q_min, q_max = box_position_limits(model)
        v_max = box_velocity_limits(model)
        tau_max = box_torque_limits(model)
        self.assertTrue(np.allclose(q_max, [+np.inf, +np.inf, 12.0, +np.inf]))
        self.assertTrue(np.allclose(q_min, [-np.inf, -np.inf, -12.0, -np.inf]))
        self.assertTrue(np.allclose(v_max, [+np.inf, 42.0, +np.inf]))
        self.assertTrue(np.allclose(tau_max, [+np.inf, 1.0, +np.inf]))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Test robot state and its randomization."""

import unittest

import numpy as np

from upkie.utils.robot_state import RobotState
from upkie.utils.robot_state_randomization import RobotStateRandomization


class TestRobotState(unittest.TestCase):
    def test_robot_state(self):
        init_state = RobotState(
            randomization=RobotStateRandomization(roll=0.0, pitch=0.0)
        )
        zyx_angles = init_state.sample_orientation(np.random).as_euler("ZYX")
        self.assertTrue(np.allclose(zyx_angles, np.zeros(3)))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Balancing using proportional control from base pitch to wheel torques."""

import gymnasium as gym
import numpy as np

import upkie.envs
from upkie.observers.base_pitch import compute_base_pitch_from_imu

upkie.envs.register()

GAIN = 10.0  # base pitch to wheel torque, in [N] * [m] / [rad]

if __name__ == "__main__":
    with gym.make("UpkieServos-v3", frequency=200.0) as env:
        action = env.get_default_action()

        # Stiff leg joints in zero configuration
        for leg_joint in ("left_hip", "left_knee", "right_hip", "right_knee"):
            action[leg_joint]["position"] = 0.0
            action[leg_joint]["velocity"] = 0.0

        # Disable position and velocity feedback in the wheels
        for wheel in ("left_wheel", "right_wheel"):
            action[wheel]["position"] = np.nan
            action[wheel]["velocity"] = 0.0
            action[wheel]["kp_scale"] = 0.0
            action[wheel]["kd_scale"] = 0.0

        obs, _ = env.reset()  # connects to the spine
        for step in range(1_000_000):
            pitch = compute_base_pitch_from_imu(obs["imu"]["orientation"])
            action["left_wheel"]["feedforward_torque"] = +GAIN * pitch
            action["right_wheel"]["feedforward_torque"] = -GAIN * pitch
            obs, _, terminated, truncated, _ = env.step(action)
            if terminated or truncated:
                obs, _ = env.reset()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

from .base_pitch import (
    compute_base_angular_velocity_from_imu,
    compute_base_pitch_from_imu,
    compute_pitch_frame_in_parent,
)

__all__ = [
    "compute_base_angular_velocity_from_imu",
    "compute_base_pitch_from_imu",
    "compute_pitch_frame_in_parent",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""!
Domain randomization of robot states.
"""

from dataclasses import dataclass
from typing import Optional

import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation as ScipyRotation


@dataclass
class RobotStateRandomization:

    """!
    Domain randomization parameters for Upkie's state.
    """

    roll: float = 0.0
    pitch: float = 0.0
    x: float = 0.0
    z: float = 0.0
    omega_x: float = 0.0
    omega_y: float = 0.0
    v_x: float = 0.0
    v_z: float = 0.0

    def update(
        self,
        roll: Optional[float] = None,
        pitch: Optional[float] = None,
        x: Optional[float] = None,
        z: Optional[float] = None,
        omega_x: Optional[float] = None,
        omega_y: Optional[float] = None,
        v_x: Optional[float] = None,
        v_z: Optional[float] = None,
    ) -> None:
        """!
        Update some fields.

        @param roll Roll angle for the rotation from the randomized frame to
            the parent frame (Euler Z-Y-X convention).
        @param pitch Pitch angle for the rotation from the randomized frame to
            the parent frame (Euler Z-Y-X convention).
        @param x Translation of the randomized frame along the x-axis of the
            parent frame.
        @param z Translation of the randomized frame along the z-axis of the
            parent frame.
        @param omega_x Angular velocity from the randomized frame to the parent
            frame, expressed in the randomized frame, along the x-axis.
        @param omega_y Angular velocity from the randomized frame to the parent
            frame, expressed in the randomized frame, along the y-axis.
        @param v_x Linear velocity from the randomized frame to the parent
            frame, expressed in the parent frame and along the x-axis.
        @param v_z Linear velocity from the randomized frame to the parent
            frame, expressed in the parent frame and along the z-axis.
        """
        if roll is not None:
            self.roll = roll
        if pitch is not None:
            self.pitch = pitch
        if x is not None:
            self.x = x
        if z is not None:
            self.z = z
        if v_x is not None:
            self.v_x = v_x
        if v_z is not None:
            self.v_z = v_z
        if omega_x is not None:
            self.omega_x = omega_x
        if omega_y is not None:
            self.omega_y = omega_y

    def sample_orientation(self, np_random) -> ScipyRotation:
        yaw_pitch_roll_bounds = np.array([0.0, self.pitch, self.roll])
        yaw_pitch_roll = np_random.uniform(
            low=-yaw_pitch_roll_bounds,
            high=+yaw_pitch_roll_bounds,
            size=3,
        )
        return ScipyRotation.from_euler("ZYX", yaw_pitch_roll)

    def sample_position(self, np_random) -> NDArray[float]:
        return np_random.uniform(
            low=np.array([-self.x, 0.0, 0.0]),
            high=np.array([+self.x, 0.0, self.z]),
            size=3,
        )

    def sample_angular_velocity(self, np_random) -> NDArray[float]:
        return np_random.uniform(
            low=np.array([-self.omega_x, -self.omega_y, 0.0]),
            high=np.array([+self.omega_x, +self.omega_y, 0.0]),
            size=3,
        )

    def sample_linear_velocity(self, np_random) -> NDArray[float]:
        return np_random.uniform(
            low=np.array([-self.v_x, 0.0, -self.v_z]),
            high=np.array([+self.v_x, 0.0, +self.v_z]),
            size=3,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""!
Robot state with optional randomization.
"""

from typing import Optional

import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation as ScipyRotation

from upkie.utils.robot_state_randomization import RobotStateRandomization


class RobotState:

    """!
    Robot state (configuration and velocity) with optional randomization.
    """

    angular_velocity_base_in_base: NDArray[float]
    joint_configuration: NDArray[float]
    joint_velocity: NDArray[float]
    linear_velocity_base_to_world_in_world: NDArray[float]
    orientation_base_in_world: ScipyRotation
    position_base_in_world: NDArray[float]
    randomization: RobotStateRandomization

    def __init__(
        self,
        angular_velocity_base_in_base: Optional[NDArray[float]] = None,
        joint_configuration: Optional[NDArray[float]] = None,
        joint_velocity: Optional[NDArray[float]] = None,
        linear_velocity_base_to_world_in_world: Optional[
            NDArray[float]
        ] = None,
        orientation_base_in_world: Optional[ScipyRotation] = None,
        position_base_in_world: Optional[NDArray[float]] = None,
        randomization: Optional[RobotStateRandomization] = None,
    ):
        self.angular_velocity_base_in_base = (
            angular_velocity_base_in_base
            if angular_velocity_base_in_base is not None
            else np.zeros(3)
        )
        self.joint_configuration = (
            joint_configuration
            if joint_configuration is not None
            else np.zeros(6)
        )
        self.joint_velocity = (
            joint_velocity if joint_velocity is not None else np.zeros(6)
        )
        self.linear_velocity_base_to_world_in_world = (
            linear_velocity_base_to_world_in_world
            if linear_velocity_base_to_world_in_world is not None
            else np.zeros(3)
        )
        self.orientation_base_in_world = (
            orientation_base_in_world
            if orientation_base_in_world is not None
            else ScipyRotation.identity()
        )
        self.position_base_in_world = (
            position_base_in_world
            if position_base_in_world is not None
            else np.array([0.0, 0.0, 0.6])  # Upkie above horizontal plane
        )
        self.randomization = (
            randomization
            if randomization is not None
            else RobotStateRandomization()
        )

    def sample_angular_velocity(self, np_random) -> NDArray[float]:
        angular_velocity_rand_to_base_in_base = (
            self.randomization.sample_angular_velocity(np_random)
        )
        angular_velocity_base_to_world_in_base = (
            self.angular_velocity_base_in_base
        )
        return (
            angular_velocity_base_to_world_in_base
            + angular_velocity_rand_to_base_in_base
        )

    def sample_linear_velocity(self, np_random) -> NDArray[float]:
        linear_velocity_rand_to_world_in_world = (
            self.randomization.sample_linear_velocity(np_random)
        )
        return (
            self.linear_velocity_base_to_world_in_world
            + linear_velocity_rand_to_world_in_world
        )

    def sample_orientation(self, np_random) -> ScipyRotation:
        rotation_rand_to_base = self.randomization.sample_orientation(
            np_random
        )
        rotation_base_to_world = self.orientation_base_in_world
        return rotation_base_to_world * rotation_rand_to_base

    def sample_position(self, np_random) -> NDArray[float]:
        position_rand_in_world = self.randomization.sample_position(np_random)
        return self.position_base_in_world + position_rand_in_world



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

import argparse
import logging
import os
import signal
import traceback
from os import path

import gin
import yaml
from loop_rate_limiters import RateLimiter
from rules_python.python.runfiles import runfiles
from servo_controller import ServoController
from vulp.spine import SpineInterface


class CompilationModeError(Exception):
    """Raised when the example is called with unexpected parameters."""


def parse_command_line_arguments() -> argparse.Namespace:
    """
    Parse command line arguments.

    Returns:
        Command-line arguments.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--show",
        default=False,
        action="store_true",
        help="show simulator while running the controller",
    )
    return parser.parse_args()


def run(
    spine: SpineInterface,
    spine_config: dict,
    frequency: float = 200.0,
) -> None:
    """!
    Read observations and send actions to the spine.

    @param spine Interface to the spine.
    @param spine_config Spine configuration dictionary.
    @param frequency Control frequency in Hz.
    """
    controller = ServoController()
    dt = 1.0 / frequency
    rate = RateLimiter(frequency, "controller")

    wheel_radius = controller.wheel_radius
    spine_config["wheel_odometry"] = {
        "signed_radius": {
            "left_wheel": +wheel_radius,
            "right_wheel": -wheel_radius,
        }
    }

    spine.start(spine_config)
    observation = spine.get_observation()  # pre-reset observation
    while True:
        observation = spine.get_observation()
        action = controller.cycle(observation, dt)
        spine.set_action(action)
        rate.sleep()


if __name__ == "__main__":
    agent_dir = path.dirname(__file__)
    deez_runfiles = runfiles.Create()
    spine_path = os.path.join(
        agent_dir,
        deez_runfiles.Rlocation("upkie/spines/bullet_spine"),
    )

    if "-opt" not in spine_path or "-fastbuild" in spine_path:
        raise CompilationModeError(
            "This example is meant to be called by ``bazel run -c opt`` "
            "so that the simulator performs well, but it seems the "
            'compilation mode is "fastbuild"? Go to the code and comment out '
            "this check if you know what you are doing :)"
        )

    # Gin configuration
    gin.parse_config_file(f"{agent_dir}/config/common.gin")
    gin.parse_config_file(f"{agent_dir}/config/bullet.gin")

    # Spine configuration
    with open(f"{agent_dir}/config/spine.yaml", "r") as fh:
        config = yaml.safe_load(fh)

    pid = os.fork()
    if pid == 0:  # child process: spine
        spine_argv = ["--spine-frequency", "1000.0"]
        args = parse_command_line_arguments()
        if args.show:
            spine_argv.append("--show")
        os.execvp(spine_path, ["bullet"] + spine_argv)
    else:
        spine = None
        try:
            spine = SpineInterface(retries=10)
            run(spine, config)
        except KeyboardInterrupt:
            logging.info("Caught a keyboard interrupt")
        except Exception:
            logging.error("Controller raised an exception")
            print("")
            traceback.print_exc()
            print("")
        finally:
            if spine:
                spine.stop()
            os.kill(pid, signal.SIGINT)  # interrupt spine child process
            os.waitpid(pid, 0)  # wait for spine to terminate



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

import argparse
import socket
import traceback
from os import path

import gin
import yaml
from loop_rate_limiters import RateLimiter
from servo_controller import ServoController
from vulp.spine import SpineInterface

from upkie.utils.raspi import configure_agent_process, on_raspi
from upkie.utils.spdlog import logging


def parse_command_line_arguments() -> argparse.Namespace:
    """
    Parse command line arguments.

    Returns:
        Command-line arguments.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-c",
        "--config",
        metavar="config",
        help="Agent configuration to apply",
        type=str,
        required=True,
    )
    return parser.parse_args()


def run(
    spine: SpineInterface,
    spine_config: dict,
    frequency: float = 200.0,
) -> None:
    """!
    Read observations and send actions to the spine.

    @param spine Interface to the spine.
    @param spine_config Spine configuration dictionary.
    @param frequency Control frequency in Hz.
    """
    controller = ServoController()
    debug: dict = {}
    dt = 1.0 / frequency
    rate = RateLimiter(frequency, "controller")

    wheel_radius = controller.wheel_radius
    spine_config["wheel_odometry"] = {
        "signed_radius": {
            "left_wheel": +wheel_radius,
            "right_wheel": -wheel_radius,
        }
    }

    spine.start(spine_config)
    observation = spine.get_observation()  # pre-reset observation
    while True:
        observation = spine.get_observation()
        action = controller.cycle(observation, dt)
        spine.set_action(action)
        debug["rate"] = {"slack": rate.slack}
        rate.sleep()


def load_gin_configuration(base_dir: str, path: str) -> None:
    logging.info(f"Loading configuration '{path}'")
    try:
        gin.parse_config_file(f"{base_dir}/{path}")
    except OSError as e:
        raise FileNotFoundError(f"Configuration '{path}' not found") from e


if __name__ == "__main__":
    if on_raspi():
        configure_agent_process()

    args = parse_command_line_arguments()
    agent_dir = path.dirname(__file__)

    # Load gin configuration files
    load_gin_configuration(agent_dir, "config/common.gin")
    if args.config == "hostname":
        hostname = socket.gethostname().lower()
        logging.info(f"Loading configuration from hostname '{hostname}'")
        load_gin_configuration(agent_dir, f"config/{hostname}.gin")
    elif args.config is not None:
        load_gin_configuration(agent_dir, f"config/{args.config}.gin")

    # Spine configuration
    with open(f"{agent_dir}/config/spine.yaml", "r") as fh:
        spine_config = yaml.safe_load(fh)

    spine = SpineInterface()
    try:
        run(spine, spine_config)
    except KeyboardInterrupt:
        logging.info("Caught a keyboard interrupt")
    except Exception:
        logging.error("Controller raised an exception")
        print("")
        traceback.print_exc()
        print("")

    logging.info("Stopping the spine...")
    try:
        spine.stop()
    except Exception:
        logging.error("Error while stopping the spine!")
        print("")
        traceback.print_exc()
        print("")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test NoisifyAction wrapper."""

import unittest

import gymnasium
import numpy as np

from upkie.envs.wrappers.low_pass_filter_action import LowPassFilterAction
from upkie.envs.wrappers.tests.envs import ActionObserverEnv


class LowPassFilterActionTestCase(unittest.TestCase):
    def test_lpf(self):
        env = ActionObserverEnv()
        lpf_env = LowPassFilterAction(env, time_constant=1.0)
        action = np.array([1.0])
        inner_action, _, _, _, _ = lpf_env.step(action)
        self.assertTrue(np.allclose(action * env.dt, inner_action))

    def test_check_env(self):
        try:
            from stable_baselines3.common.env_checker import check_env

            env = gymnasium.make("Pendulum-v1")
            lpf_env = LowPassFilterAction(env, time_constant=1.0)
            check_env(lpf_env)
        except ImportError:
            pass


if __name__ == "__main__":
    unittest.main()  # necessary for `bazel test`



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test UpkieGroundVelocity."""

import unittest

import numpy as np
import posix_ipc

from upkie.envs import UpkieGroundVelocity
from upkie.envs.tests.mock_spine import MockSpine


class TestUpkieGroundVelocity(unittest.TestCase):
    def setUp(self):
        shm_name = "/vroum"
        shared_memory = posix_ipc.SharedMemory(
            shm_name, posix_ipc.O_RDWR | posix_ipc.O_CREAT, size=42
        )
        self.env = UpkieGroundVelocity(
            fall_pitch=1.0,
            frequency=100.0,
            max_ground_velocity=1.0,
            shm_name=shm_name,
        )
        shared_memory.close_fd()
        self.env._spine = MockSpine()

    def test_reset(self):
        observation, info = self.env.reset()
        spine_observation = info["spine_observation"]
        self.assertAlmostEqual(
            observation[1], spine_observation["wheel_odometry"]["position"]
        )
        self.assertGreaterEqual(spine_observation["number"], 1)

    def test_reward(self):
        observation, info = self.env.reset()
        action = np.zeros(self.env.action_space.shape)
        observation, reward, terminated, truncated, _ = self.env.step(action)
        self.assertAlmostEqual(reward, 1.0)  # survival reward

    def test_check_env(self):
        try:
            from stable_baselines3.common.env_checker import check_env

            check_env(self.env)
        except ImportError:
            pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Genuflect while lying on a horizontal floor."""

import gymnasium as gym
import numpy as np
from scipy.spatial.transform import Rotation as ScipyRotation

import upkie.envs
from upkie.utils.robot_state import RobotState

NB_GENUFLECTIONS = 10
GENUFLECTION_STEPS = 200
AMPLITUDE = 1.0  # in radians

if __name__ == "__main__":
    upkie.envs.register()
    with gym.make(
        "UpkieServos-v3",
        frequency=200.0,
        init_state=RobotState(
            orientation_base_in_world=ScipyRotation.from_quat(
                [0.0, -0.707, 0.0, 0.707]  # SciPy convention: (x, y, z, w)
            ),
            position_base_in_world=np.array([0.0, 0.0, 0.1]),
        ),
    ) as env:
        action = env.get_default_action()
        observation, _ = env.reset()  # connects to the spine
        for step in range(NB_GENUFLECTIONS * GENUFLECTION_STEPS):
            x = float(step % GENUFLECTION_STEPS) / GENUFLECTION_STEPS
            y = 4.0 * x * (1.0 - x)  # in [0, 1]
            q_0134 = AMPLITUDE * y * np.array([1.0, -2.0, -1.0, 2.0])
            action["left_hip"]["position"] = q_0134[0]
            action["left_knee"]["position"] = q_0134[1]
            action["right_hip"]["position"] = q_0134[2]
            action["right_knee"]["position"] = q_0134[3]
            action["left_wheel"]["position"] = np.nan
            action["right_wheel"]["position"] = np.nan
            observation, _, _, _, _ = env.step(action)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Wheel balancing using model predictive control of an LTV system."""

import gymnasium as gym
import numpy as np
from numpy.typing import NDArray
from qpmpc import solve_mpc
from qpmpc.systems import WheeledInvertedPendulum

import upkie.envs
from upkie.utils.clamp import clamp_and_warn

upkie.envs.register()


def get_target_states(
    pendulum: WheeledInvertedPendulum,
    state: NDArray[float],
    target_vel: float = 0.0,
) -> NDArray[float]:
    """!
    Define the reference state trajectory over the receding horizon.

    @param pendulum Instance from which we read receding-horizon properties.
    @param state Initial state of the wheeled inverted pendulum system.
    @param target_vel Target ground velocity in m/s.
    @return Reference state trajectory over the horizon.
    """
    nx = WheeledInvertedPendulum.STATE_DIM
    target_states = np.zeros((pendulum.nb_timesteps + 1) * nx)
    for k in range(pendulum.nb_timesteps + 1):
        target_states[k * nx] = (
            state[0] + (k * pendulum.sampling_period) * target_vel
        )
        target_states[k * nx + 2] = target_vel
    return target_states


def balance(env: gym.Env, nb_env_steps: int = 10_000) -> None:
    """!
    Balancer Upkie by closed-loop MPC on its gym environment.

    @param env Gym environment to Upkie.

    @note This example rebuilds the QP problem at every step and does not
    implement hot-starting, both of which impact performance. See the MPC
    balancer agent for a more complete example.
    """
    pendulum = WheeledInvertedPendulum(
        length=0.3,
        max_ground_accel=10.0,
        nb_timesteps=50,
        sampling_period=0.02,
    )
    mpc_problem = pendulum.build_mpc_problem(
        terminal_cost_weight=1.0,
        stage_state_cost_weight=1e-2,
        stage_input_cost_weight=1e-3,
    )

    env.reset()  # connects to the spine
    commanded_velocity = 0.0
    action = np.zeros(env.action_space.shape)
    for step in range(nb_env_steps):
        action[0] = commanded_velocity
        observation, _, terminated, truncated, info = env.step(action)
        if terminated or truncated:
            observation, info = env.reset()
            commanded_velocity = 0.0

        theta, ground_pos, theta_dot, ground_vel = observation
        initial_state = np.array([ground_pos, theta, ground_vel, theta_dot])
        target_states = get_target_states(pendulum, initial_state)
        mpc_problem.update_initial_state(initial_state)
        mpc_problem.update_goal_state(
            target_states[-WheeledInvertedPendulum.STATE_DIM :]
        )
        mpc_problem.update_target_states(
            target_states[: -WheeledInvertedPendulum.STATE_DIM]
        )

        # NB: we solve the MPC problem "cold" here, i.e. without hot-starting.
        # This will likely take too much time to fit in a 200 Hz control loop
        # on most current computers. See the MPC balancer for a more performant
        # implementation that runs in real-time on a Raspberry Pi 4 Model B.
        plan = solve_mpc(mpc_problem, solver="proxqp")
        if not plan.is_empty:
            commanded_accel = plan.first_input
            commanded_velocity = clamp_and_warn(
                commanded_velocity + commanded_accel * env.dt / 2.0,
                lower=-1.0,
                upper=+1.0,
                label="commanded_velocity",
            )


if __name__ == "__main__":
    with gym.make("UpkieGroundVelocity-v3", frequency=200.0) as env:
        balance(env),



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test UpkieBaseEnv."""

import unittest

import numpy as np
import posix_ipc
from gymnasium import spaces
from numpy.typing import NDArray

from upkie.envs import UpkieBaseEnv
from upkie.envs.tests.mock_spine import MockSpine


class UpkieTestEnv(UpkieBaseEnv):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.observation_space = spaces.Box(
            -1.0,
            +1.0,
            shape=(1,),
            dtype=np.float32,
        )
        self.action_space = spaces.Box(
            -1.0,
            1.0,
            shape=(1,),
            dtype=np.float32,
        )

    def parse_first_observation(self, spine_observation: dict) -> None:
        pass

    def get_env_observation(self, spine_observation: dict) -> NDArray[float]:
        return np.full((1,), 0.5, dtype=self.observation_space.dtype)

    def get_spine_action(self, action: NDArray[float]) -> dict:
        return {"test": action}

    def get_reward(
        self, observation: NDArray[float], action: NDArray[float]
    ) -> float:
        return 1.0


class TestUpkieBaseEnv(unittest.TestCase):
    def setUp(self):
        shm_name = "/vroum"
        shared_memory = posix_ipc.SharedMemory(
            shm_name, posix_ipc.O_RDWR | posix_ipc.O_CREAT, size=42
        )
        self.env = UpkieTestEnv(
            fall_pitch=1.0,
            frequency=100.0,
            shm_name=shm_name,
            spine_config=None,
        )
        shared_memory.close_fd()
        self.env._spine = MockSpine()

    def test_reset(self):
        _, info = self.env.reset()
        spine_observation = info["spine_observation"]
        self.assertGreaterEqual(spine_observation["number"], 1)

    def test_spine_config(self):
        """Check that runtime and default configs are merged properly."""
        shm_name = "/vroum"
        shared_memory = posix_ipc.SharedMemory(
            shm_name, posix_ipc.O_RDWR | posix_ipc.O_CREAT, size=42
        )
        env = UpkieTestEnv(
            fall_pitch=1.0,
            frequency=100.0,
            shm_name=shm_name,
            spine_config={"some_value": 12, "bullet": {"gui": False}},
        )
        shared_memory.close_fd()
        self.assertEqual(env._spine_config["some_value"], 12)
        self.assertEqual(env._spine_config["some_value"], 12)
        self.assertEqual(env._spine_config["bullet"]["gui"], False)

    def test_check_env(self):
        try:
            from stable_baselines3.common.env_checker import check_env

            check_env(self.env)
        except ImportError:
            pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

import gin
import numpy as np
from wheel_controller import WheelController

from upkie.utils.clamp import clamp


@gin.configurable
class ServoController:
    """!
    Balance Upkie using its wheels.
    """

    gain_scale: float
    turning_gain_scale: float

    def __init__(
        self,
        gain_scale: float,
        turning_gain_scale: float,
        wheel_distance: float,
    ):
        """!
        Create controller.

        @param gain_scale PD gain scale for hip and knee joints.
        @param turning_gain_scale Additional gain scale added when the robot is
            turning to keep the legs stiff in spite of the ground pulling them
            apart.
        @param wheel_distance Lateral distance between the two wheels in
            meters. This controller does not handle the case where the two
            wheels are not in the lateral plane.
        """
        ## PD gain scale for hip and knee joints.
        self.gain_scale = clamp(gain_scale, 0.1, 2.0)

        ## Translation from the left contact frame to the right contact frame,
        ## expressed in the left contact frame.
        self.position_right_in_left = np.array([0.0, wheel_distance, 0.0])

        ## Additional gain scale added when the robot is turning to keep the
        ## legs stiff while the ground pulls them apart.
        self.turning_gain_scale = turning_gain_scale

        self.__servo_action = None
        self.__wheel_balancer = WheelController()  # type: ignore

    @property
    def wheel_radius(self) -> float:
        """!Wheel radius in [m]."""
        return self.__wheel_balancer.wheel_radius

    def initialize_servo_action(self, observation: dict) -> None:
        """!
        Initialize default servo action from initial observation.

        @param observation Initial observation.
        """
        self.__servo_action = {
            joint: {
                "position": observation["servo"][joint]["position"],
                "velocity": 0.0,
                "maximum_torque": 10.0,  # qdd100 actuators
            }
            for joint in (
                f"{side}_{func}"
                for side in ("left", "right")
                for func in ("hip", "knee")
            )
        }
        self.__servo_action.update(
            {
                wheel: {
                    "position": np.nan,
                    "velocity": 0.0,
                    "maximum_torque": 1.0,  # mj5208 actuators
                }
                for wheel in ("left_wheel", "right_wheel")
            }
        )

    def cycle(self, observation: dict, dt: float) -> dict:
        """!
        Compute action for a new cycle.

        @param observation Latest observation.
        @param dt Duration in seconds until next cycle.
        @return Dictionary with the new action and some internal state for
            logging.
        """
        if self.__servo_action is None:
            self.initialize_servo_action(observation)

        # Compute wheel velocities for balancing
        self.__wheel_balancer.cycle(observation, dt)
        wheel_velocities = self.__wheel_balancer.get_wheel_velocities(
            self.position_right_in_left
        )
        left_wheel_velocity, right_wheel_velocity = wheel_velocities
        self.__servo_action["left_wheel"]["velocity"] = left_wheel_velocity
        self.__servo_action["right_wheel"]["velocity"] = right_wheel_velocity

        # Increase leg stiffness while turning
        turning_prob = self.__wheel_balancer.turning_probability
        kp_scale = self.gain_scale + self.turning_gain_scale * turning_prob
        kd_scale = self.gain_scale + self.turning_gain_scale * turning_prob
        for joint_name in ["left_hip", "left_knee", "right_hip", "right_knee"]:
            self.__servo_action[joint_name]["kp_scale"] = kp_scale
            self.__servo_action[joint_name]["kd_scale"] = kd_scale

        return {
            "servo": self.__servo_action,
            "wheel_balancer": self.__wheel_balancer.log(),
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

import gymnasium
import numpy as np
from gymnasium import spaces
from gymnasium.wrappers import FrameStack, RescaleAction
from settings import EnvSettings

from upkie.envs import UpkieGroundVelocity
from upkie.envs.wrappers import (
    AddActionToObservation,
    DifferentiateAction,
    LowPassFilterAction,
    NoisifyAction,
    NoisifyObservation,
)


def make_training_env(
    velocity_env: UpkieGroundVelocity,
    env_settings: EnvSettings,
) -> gymnasium.Wrapper:
    action_noise = np.array(env_settings.action_noise)
    observation_noise = np.array(env_settings.observation_noise)
    noisy_obs_env = NoisifyObservation(velocity_env, noise=observation_noise)
    noisy_env = NoisifyAction(noisy_obs_env, noise=action_noise)
    filtered_env = LowPassFilterAction(
        noisy_env,
        time_constant=spaces.Box(*env_settings.action_lpf),
    )
    return filtered_env


def make_accel_env(
    velocity_env: UpkieGroundVelocity,
    env_settings: EnvSettings,
    training: bool,
) -> gymnasium.Wrapper:
    inner_env = (
        make_training_env(velocity_env, env_settings)
        if training
        else velocity_env
    )
    history_env = FrameStack(
        AddActionToObservation(inner_env),
        env_settings.history_size,
    )
    accel_env = DifferentiateAction(
        history_env,
        min_derivative=-env_settings.max_ground_accel,
        max_derivative=+env_settings.max_ground_accel,
    )
    rescaled_accel_env = RescaleAction(
        accel_env,
        min_action=-1.0,
        max_action=+1.0,
    )
    return rescaled_accel_env


def make_ppo_balancer_env(
    velocity_env: UpkieGroundVelocity,
    env_settings: EnvSettings,
    training: bool,
) -> gymnasium.Wrapper:
    return make_accel_env(velocity_env, env_settings, training=training)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria
#
# This file incorporates work covered by the following copyright and
# permission notice:
#
#     setup-system.py from github.com:mjbots/quad
#     Copyright 2018-2020 Josh Pieper
#     SPDX-License-Identifier: Apache-2.0


"""Set up a Raspberry Pi 4b to run software for an Upkie wheeled biped.

The base operating system should already be installed. This script is intended
to be run as root, like: ``sudo ./setup-raspi.py``.
"""

import datetime
import os
import subprocess
import sys
import time

ORIG_SUFFIX = time.strftime(".orig-%Y%m%d-%H%M%S")

# Log each setup step, in case
# ============================

logging_depth = 0


def log_message(message: str, indent: int = 0) -> None:
    """Log a single message.

    Args:
        message: Message to log.
        indent: Indentation level.
    """
    logging_indent = " | " * (logging_depth + indent)
    now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    print(f"{now}: {logging_indent}{message}")


def log_method(func):
    """Decorator to log function calls."""

    def wrapped_function(*args, **kwargs):
        global logging_depth
        try:
            logging_depth += 1
            args_repr = (
                f'"{args[0]}"' + (", ..." if len(args) > 1 else "")
                if args
                else ""
            )
            log_message(f"{func.__name__}({args_repr})", indent=-1)
            result = func(*args, **kwargs)
            return result
        finally:
            logging_depth -= 1

    return wrapped_function


# Utility functions
# =================


@log_method
def run(*args, **kwargs):
    """Run a subprocess."""
    subprocess.check_call(*args, shell=True, **kwargs)


@log_method
def install_packages():
    """Install Debian and PyPI packages."""
    run("apt-get install --yes python3-pip")  # for moteus-pi3hat
    run("apt-get install --yes screen vim")  # convenient later on
    run("pip install moteus-pi3hat")
    run("pip install upkie")


@log_method
def configure_cpu_isolation(filename="/boot/cmdline.txt"):
    """Make sure CPU isolation is configured.

    Args:
        filename: Path to the boot cmdline configuration file.
    """
    keyword, value = "isolcpus", "1,2,3"
    file_content = [
        x.strip() for x in open(filename, encoding="utf-8").readlines()
    ]
    assert len(file_content) == 1

    new_item = "{}={}".format(keyword, value)
    items = [x.strip() for x in file_content[0].split(" ")]
    present = [x for x in items if x == keyword or x.startswith(keyword + "=")]
    if len(present) > 0:
        new_items = [
            x
            if not (x == keyword or x.startswith(keyword + "="))
            else new_item
            for x in items
        ]
    else:  # len(present) == 0
        new_items = items + [new_item]

    if new_items == items:
        log_message("configure_cpu_isolation(): Already configured")
        return

    log_message(
        "configure_cpu_isolation(): Adding {}={} to {}".format(
            keyword, value, filename
        )
    )

    with open(filename, "w", encoding="utf-8") as f:
        f.write(" ".join(new_items) + "\n")


if __name__ == "__main__":
    if os.geteuid() != 0:
        args = ["sudo", "-E", sys.executable] + sys.argv + [os.environ]
        os.execlpe("sudo", *args)
    install_packages()
    configure_cpu_isolation()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

from typing import Optional

from .spdlog import logging


def clamp(
    value: float, lower: Optional[float] = None, upper: Optional[float] = None
) -> float:
    """!
    Clamp a value between a lower and an upper bound.

    @param value Value to clamp.
    @param lower Lower bound.
    @param upper Upper bound.
    """
    if lower is not None and value < lower:
        return lower
    if upper is not None and value > upper:
        return upper
    return value


def clamp_abs(value: float, bound: float):
    """!
    Clamp a value's absolute value, keeping its sign.

    @param value Value to clamp.
    @param bound Absolute value bound.
    """
    return clamp(value, -bound, bound)


def clamp_and_warn(value: float, lower: float, upper: float, label: str):
    """!
    Clamp a value between a lower and an upper bound, warning if the value is
    changed.

    @param value Value to clamp.
    @param lower Lower bound.
    @param upper Upper bound.
    @param label Label to describe the value.
    """
    if value < lower:
        logging.warning(f"{label}={value} clamped to {lower=}")
        return lower
    elif value > upper:
        logging.warning(f"{label}={value} clamped to {upper=}")
        return upper
    return value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

from typing import Optional, Tuple

import numpy as np
from numpy.typing import NDArray

from upkie.utils.clamp import clamp
from upkie.utils.rotations import rotation_matrix_from_quaternion


def compute_pitch_frame_in_parent(
    orientation_frame_in_parent: NDArray[float],
) -> float:
    """!
    Get pitch angle of a given frame relative to the parent vertical.

    Figure:

    ```
        parent z
          ^    frame z
          |     /
          |    /
          |   /
          |  /
          | /
          |/
         (x)-----------> heading vector
           \\  )
            \\+  positive pitch
             \\
              \\
               \\
              frame x
    ```

    @param orientation_frame_in_parent Rotation matrix from the target frame to
        the parent frame.

    @returns Angle from the parent z-axis (gravity) to the frame z-axis.
        Equivalently, angle from the heading vector to the sagittal vector of
        the frame.

    @note Angle is positive in the trigonometric sense (CCW positive, CW
    negative) in the heading-vertical plane directed by the lateral vector,
    which is ``(x)`` in the above schematic (pointing away) and not ``(.)``
    (poiting from screen to the reader).
    """
    sagittal = orientation_frame_in_parent[:, 0]
    sagittal /= np.linalg.norm(sagittal)  # needed for prec around cos_pitch=0.
    heading_in_parent = sagittal - sagittal[2] * np.array([0.0, 0.0, 1.0])
    heading_in_parent /= np.linalg.norm(heading_in_parent)
    if orientation_frame_in_parent[2, 2] < 0:
        heading_in_parent *= -1.0
    sign = +1 if sagittal[2] < 0 else -1
    cos_pitch = clamp(np.dot(sagittal, heading_in_parent), -1.0, 1.0)
    pitch = float(sign * np.arccos(cos_pitch))  # float rather than np.float64
    return pitch


def compute_base_orientation_from_imu(
    quat_imu_in_ars: Tuple[float, float, float, float],
    rotation_base_to_imu: Optional[NDArray[float]] = None,
) -> NDArray[float]:
    """!
    Get the orientation of the base frame with respect to the world frame.

    @param quat_imu_in_ars Quaternion representing the rotation matrix from the
        IMU frame to the  attitude reference system (ARS) frame, in ``[w, x, y,
        z]`` format.
    @param rotation_base_to_imu Rotation matrix from the base frame to the IMU
        frame. When not specified, the default Upkie mounting orientation is
        used.
    @returns Rotation matrix from the base frame to the world frame.
    """
    if rotation_base_to_imu is None:
        # Default Upkie mounting orientation: USB connectors of the raspi
        # pointing to the left of the robot (XT-30 of the pi3hat to the right)
        rotation_base_to_imu = np.diag([-1.0, 1.0, -1.0])

    rotation_imu_to_ars = rotation_matrix_from_quaternion(quat_imu_in_ars)

    # The attitude reference system frame has +x forward, +y right and +z down,
    # whereas our world frame has +x forward, +y left and +z up:
    # https://github.com/mjbots/pi3hat/blob/ab632c82bd501b9fcb6f8200df0551989292b7a1/docs/reference.md#orientation
    rotation_ars_to_world = np.diag([1.0, -1.0, -1.0])

    rotation_base_to_world = (
        rotation_ars_to_world @ rotation_imu_to_ars @ rotation_base_to_imu
    )
    return rotation_base_to_world


def compute_base_pitch_from_imu(
    quat_imu_in_ars: Tuple[float, float, float, float],
    rotation_base_to_imu: Optional[NDArray[float]] = None,
) -> float:
    """!
    Get pitch angle of the base frame relative to the world frame.

    @param quat_imu_in_ars Quaternion representing the rotation matrix from
        the IMU frame to the  attitude reference system (ARS) frame, in ``[w,
        x, y, z]`` format.
    @param rotation_base_to_imu Rotation matrix from the base frame to the IMU
        frame. When not specified, the default Upkie mounting orientation is
        used.
    @returns Angle from the world z-axis (unit vector opposite to gravity) to
        the base z-axis. This angle is positive when the base leans forward.
    """
    rotation_base_to_world = compute_base_orientation_from_imu(
        quat_imu_in_ars, rotation_base_to_imu
    )
    pitch_base_in_world = compute_pitch_frame_in_parent(rotation_base_to_world)
    return pitch_base_in_world


def compute_base_angular_velocity_from_imu(
    angular_velocity_imu_in_imu: NDArray[float],
) -> NDArray[float]:
    r"""!
    Compute the body angular velocity of the base from IMU readings.

    @param angular_velocity_imu_in_imu Angular velocity from the IMU.
    @returns Body angular velocity of the base frame.

    Calculation checks:

    - \f$I\f$: IMU frame
    - \f$B\f$: base frame
    - \f$W\f$: world (inertial) frame

    Our input is \f${}_I \omega_{WI}\f$, we seek \f${}_B \omega_{WB}\f$.

    \f$
    \begin{align*}
        \dot{R}_{WI} & = R_{WI} ({}_I \omega_{WI} \times) \\
        R_{WB} & = R_{WI} R_{IB} \\
        \dot{R}_{WB} & = \dot{R}_{WI} R_{IB} \\
        & = R_{WI} ({}_I \omega_{WI} \times) R_{IB} \\
        & = R_{WI} R_{IB} ({}_B \omega_{WI} \times) \\
        & = R_{WB} ({}_B \omega_{WI} \times) = R_{WB} ({}_B \omega_{WB} \times)
    \end{align*}
    \f$

    Thus \f${}_B \omega_{WB} = R_{BI} {}_I \omega_{WI}\f$.
    """
    # TODO(scaron): move to config
    rotation_base_to_imu = np.diag([-1.0, 1.0, -1.0])
    rotation_imu_to_base = rotation_base_to_imu.T
    angular_velocity_base_in_base = (
        rotation_imu_to_base @ angular_velocity_imu_in_imu
    )
    return angular_velocity_base_in_base



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test clamping utility functions."""

import unittest

import numpy as np

from upkie.utils.clamp import clamp, clamp_abs, clamp_and_warn


class TestClamp(unittest.TestCase):
    def test_clamp(self):
        self.assertAlmostEqual(clamp(-2.1, -2.0, 2.0), -2.0)
        self.assertAlmostEqual(clamp(3.1, -2.0, 2.0), 2.0)

    def test_clamp_abs(self):
        self.assertAlmostEqual(clamp_abs(-2.1, 2.0), -2.0)
        self.assertAlmostEqual(clamp_abs(3.1, 2.0), 2.0)

    def test_clamp_and_warn(self):
        self.assertAlmostEqual(clamp_and_warn(-2.1, -2.0, 2.0, "x"), -2.0)
        self.assertAlmostEqual(clamp_and_warn(3.1, -2.0, 2.0, "x"), 2.0)

    def test_clamp_inf(self, x=1.42):
        self.assertAlmostEqual(clamp(x, -np.inf, np.inf), x)
        self.assertAlmostEqual(clamp_abs(x, np.inf), x)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

import math
from dataclasses import dataclass
from typing import Dict, Optional, Tuple

import numpy as np
from gymnasium import spaces
from numpy.typing import NDArray

from upkie.observers.base_pitch import (
    compute_base_angular_velocity_from_imu,
    compute_base_pitch_from_imu,
)
from upkie.utils.exceptions import UpkieException
from upkie.utils.filters import low_pass_filter
from upkie.utils.robot_state import RobotState

from .upkie_base_env import UpkieBaseEnv


class UpkieGroundVelocity(UpkieBaseEnv):

    """!
    Environment where Upkie is used as a wheeled inverted pendulum.

    The environment id is ``UpkieGroundVelocity-v3``. Model assumptions are
    summarized in the <a
    href="https://scaron.info/robotics/wheeled-inverted-pendulum-model.html">following
    note</a>.

    @note For reinforcement learning with neural networks: the observation
    space and action space are not normalized.

    ### Action space

    The action corresponds to the ground velocity resulting from wheel
    velocities. The action vector is simply:

    <table>
        <tr>
            <td><strong>Index</strong></td>
            <td><strong>Description</strong></td>
            </tr>
        <tr>
            <td>``0``</td>
            <td>Ground velocity in [m] / [s].</td>
        </tr>
    </table>

    Note that, while this action is not normalized, [-1, 1] m/s is a reasonable
    range for ground velocities.

    ### Observation space

    Vectorized observations have the following structure:

    <table>
        <tr>
            <td><strong>Index</strong></td>
            <td><strong>Description</strong></td>
        </tr>
        <tr>
            <td>0</td>
            <td>Pitch angle of the base with respect to the world vertical, in
            radians. This angle is positive when the robot leans forward.</td>
        </tr>
        <tr>
            <td>1</td>
            <td>Position of the average wheel contact point, in meters.</td>
        </tr>
        <tr>
            <td>2</td>
            <td>Body angular velocity of the base frame along its lateral axis,
            in radians per seconds.</td>
        </tr>
        <tr>
            <td>3</td>
            <td>Velocity of the average wheel contact point, in meters per
            seconds.</td>
        </tr>
    </table>

    ### Attributes

    The environment class defines the following attributes:

    - ``leg_return_period``: Time constant for the legs (hips and knees) to
        revert to their neutral configuration.
    - ``version``: Environment version number.
    - ``wheel_radius``: Wheel radius in [m].

    """

    LEG_JOINTS = [
        f"{side}_{joint}"
        for side in ("left", "right")
        for joint in ("hip", "knee")
    ]

    @dataclass
    class RewardWeights:
        position: float = 1.0
        velocity: float = 1.0

    version: int = 3
    wheel_radius: float

    def __init__(
        self,
        fall_pitch: float = 1.0,
        frequency: float = 200.0,
        init_state: Optional[RobotState] = None,
        leg_return_period: float = 1.0,
        max_ground_velocity: float = 1.0,
        regulate_frequency: bool = True,
        reward_weights: Optional[RewardWeights] = None,
        shm_name: str = "/vulp",
        spine_config: Optional[dict] = None,
        wheel_radius: float = 0.06,
    ):
        """!
        Initialize environment.

        @param fall_pitch Fall detection pitch angle, in radians.
        @param frequency Regulated frequency of the control loop, in Hz.
        @param init_state Initial state of the robot, only used in simulation.
        @param leg_return_period Time constant for the legs (hips and knees) to
            revert to their neutral configuration.
        @param max_ground_velocity Maximum commanded ground velocity in m/s.
        @param regulate_frequency Enables loop frequency regulation.
        @param reward_weights Coefficients before each reward term.
        @param shm_name Name of shared-memory file.
        @param spine_config Additional spine configuration overriding the
            defaults from ``//config:spine.yaml``. The combined configuration
            dictionary is sent to the spine at every :func:`reset`.
        @param wheel_radius Wheel radius in [m].
        """
        super().__init__(
            fall_pitch=fall_pitch,
            frequency=frequency,
            init_state=init_state,
            regulate_frequency=regulate_frequency,
            shm_name=shm_name,
            spine_config=spine_config,
        )

        if self.dt is None:
            raise UpkieException("This environment needs a loop frequency")

        reward_weights: UpkieGroundVelocity.RewardWeights = (
            reward_weights
            if reward_weights is not None
            else UpkieGroundVelocity.RewardWeights()
        )

        # gymnasium.Env: observation_space
        MAX_BASE_PITCH: float = np.pi
        MAX_GROUND_POSITION: float = float("inf")
        MAX_BASE_ANGULAR_VELOCITY: float = 1000.0  # rad/s
        observation_limit = np.array(
            [
                MAX_BASE_PITCH,
                MAX_GROUND_POSITION,
                MAX_BASE_ANGULAR_VELOCITY,
                max_ground_velocity,
            ],
            dtype=float,
        )
        self.observation_space = spaces.Box(
            -observation_limit,
            +observation_limit,
            shape=observation_limit.shape,
            dtype=observation_limit.dtype,
        )

        # gymnasium.Env: action_space
        action_limit = np.array([max_ground_velocity], dtype=float)
        self.action_space = spaces.Box(
            -action_limit,
            +action_limit,
            shape=action_limit.shape,
            dtype=action_limit.dtype,
        )

        self.__leg_servo_action = {
            joint: {
                "position": None,
                "velocity": 0.0,
                "maximum_torque": 10.0,  # qdd100 actuators
            }
            for joint in self.LEG_JOINTS
        }

        self.leg_return_period = leg_return_period
        self.reward_weights = reward_weights
        self.wheel_radius = wheel_radius

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[NDArray[float], Dict]:
        """!
        Resets the environment and get an initial observation.

        @param seed Number used to initialize the environment’s internal random
            number generator.
        @param options Currently unused.
        @returns
            - ``observation``: Initial vectorized observation, i.e. an element
              of the environment's ``observation_space``.
            - ``info``: Dictionary with auxiliary diagnostic information. For
              Upkie this is the full observation dictionary sent by the spine.
        """
        return super().reset(seed=seed)

    def parse_first_observation(self, spine_observation: dict) -> None:
        """!
        Parse first observation after the spine interface is initialized.

        @param spine_observation First observation.
        """
        for joint in self.LEG_JOINTS:
            position = spine_observation["servo"][joint]["position"]
            self.__leg_servo_action[joint]["position"] = position

    def get_env_observation(self, spine_observation: dict) -> NDArray[float]:
        """!
        Extract environment observation from spine observation dictionary.

        @param spine_observation Spine observation dictionary.
        @returns Environment observation vector.
        """
        imu = spine_observation["imu"]
        pitch_base_in_world = compute_base_pitch_from_imu(imu["orientation"])
        angular_velocity_base_in_base = compute_base_angular_velocity_from_imu(
            spine_observation["imu"]["angular_velocity"]
        )
        ground_position = spine_observation["wheel_odometry"]["position"]
        ground_velocity = spine_observation["wheel_odometry"]["velocity"]

        obs = np.empty(4, dtype=float)
        obs[0] = pitch_base_in_world
        obs[1] = ground_position
        obs[2] = angular_velocity_base_in_base[1]
        obs[3] = ground_velocity
        return obs

    def get_leg_servo_action(self) -> Dict[str, Dict[str, float]]:
        """!
        Get servo actions for both hip and knee joints.

        @returns Servo action dictionary.
        """
        for joint in self.LEG_JOINTS:
            prev_position = self.__leg_servo_action[joint]["position"]
            new_position = low_pass_filter(
                prev_output=prev_position,
                cutoff_period=self.leg_return_period,
                new_input=0.0,
                dt=self.dt,
            )
            self.__leg_servo_action[joint]["position"] = new_position
        return self.__leg_servo_action.copy()

    def get_spine_action(self, action: NDArray[float]) -> dict:
        """!
        Convert environment action to a spine action dictionary.

        @param action Environment action.
        @returns Spine action dictionary.
        """
        ground_velocity = action[0]
        wheel_velocity = ground_velocity / self.wheel_radius
        servo_dict = self.get_leg_servo_action()
        servo_dict.update(
            {
                "left_wheel": {
                    "position": math.nan,
                    "velocity": +wheel_velocity,
                    "maximum_torque": 1.0,  # mj5208 actuator
                },
                "right_wheel": {
                    "position": math.nan,
                    "velocity": -wheel_velocity,
                    "maximum_torque": 1.0,  # mj5208 actuator
                },
            }
        )
        spine_action = {"servo": servo_dict}
        return spine_action

    def get_reward(
        self,
        observation: NDArray[float],
        action: NDArray[float],
    ) -> float:
        """!
        Get reward from observation and action.

        @param observation Environment observation vector.
        @param action Environment action vector.
        @returns Reward.
        """
        pitch = observation[0]
        ground_position = observation[1]
        angular_velocity = observation[2]
        ground_velocity = observation[3]

        tip_height = 0.58  # [m]
        tip_position = ground_position + tip_height * np.sin(pitch)
        tip_velocity = (
            ground_velocity + tip_height * angular_velocity * np.cos(pitch)
        )

        std_position = 0.05  # [m]
        position_reward = np.exp(-((tip_position / std_position) ** 2))
        velocity_penalty = -abs(tip_velocity)

        return (
            self.reward_weights.position * position_reward
            + self.reward_weights.velocity * velocity_penalty
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

import argparse
import logging
import os
from typing import Tuple

import gin
import gymnasium as gym
import numpy as np
from envs import make_ppo_balancer_env
from settings import EnvSettings, PPOSettings, TrainingSettings
from stable_baselines3 import PPO

import upkie.envs
from upkie.utils.raspi import configure_agent_process, on_raspi
from upkie.utils.robot_state import RobotState
from upkie.utils.robot_state_randomization import RobotStateRandomization

upkie.envs.register()


def parse_command_line_arguments() -> argparse.Namespace:
    """!
    Parse command line arguments.

    @returns Command-line arguments.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "policy",
        nargs="?",
        help="path to the policy parameters file",
    )
    parser.add_argument(
        "--training",
        default=False,
        action="store_true",
        help="add noise and delays as in training",
    )
    return parser.parse_args()


def get_tip_state(
    observation, tip_height: float = 0.58
) -> Tuple[float, float]:
    """!
    Compute the state of the virtual tip used in the agent's reward.

    This extra info is for logging only.

    @param observation Observation vector.
    @param tip_height Height of the virtual tip.
    @returns Pair of tip (position, velocity) in the sagittal plane.
    """
    pitch = observation[0]
    ground_position = observation[1]
    angular_velocity = observation[2]
    ground_velocity = observation[3]
    tip_position = ground_position + tip_height * np.sin(pitch)
    tip_velocity = ground_velocity + tip_height * angular_velocity * np.cos(
        pitch
    )
    return tip_position, tip_velocity


def run_policy(env: gym.Wrapper, policy) -> None:
    """!
    Run the policy on a given environment.

    @param env Upkie environment, wrapped by the agent.
    @param policy MLP policy to follow.
    """
    action = np.zeros(env.action_space.shape)
    observation, info = env.reset()
    reward = 0.0
    while True:
        action, _ = policy.predict(observation, deterministic=True)
        tip_position, tip_velocity = get_tip_state(observation[-1])
        env.log(
            {
                "action": action,
                "observation": observation[-1],
                "reward": reward,
                "tip_position": tip_position,
                "tip_velocity": tip_velocity,
            }
        )
        observation, reward, terminated, truncated, info = env.step(action)
        if terminated or truncated:
            observation, info = env.reset()


def main(policy_path: str, training: bool) -> None:
    """!
    Load environment and policy, and run the latter on the former.

    @param policy_path Path to policy parameters.
    @param training If True, add training noise and domain randomization.
    """
    env_settings = EnvSettings()
    init_state = None
    if training:
        training_settings = TrainingSettings()
        init_state = RobotState(
            randomization=RobotStateRandomization(
                **training_settings.init_rand
            ),
        )
    with gym.make(
        env_settings.env_id,
        frequency=env_settings.agent_frequency,
        init_state=init_state,
        max_ground_velocity=env_settings.max_ground_velocity,
        regulate_frequency=True,
        spine_config=env_settings.spine_config,
    ) as velocity_env:
        env = make_ppo_balancer_env(
            velocity_env,
            env_settings,
            training=training,
        )
        ppo_settings = PPOSettings()
        policy = PPO(
            "MlpPolicy",
            env,
            policy_kwargs={
                "net_arch": {
                    "pi": ppo_settings.net_arch_pi,
                    "vf": ppo_settings.net_arch_vf,
                },
            },
            verbose=0,
        )
        policy.set_parameters(policy_path)
        run_policy(env, policy)


if __name__ == "__main__":
    if on_raspi():
        configure_agent_process()

    agent_dir = os.path.abspath(os.path.dirname(__file__))
    args = parse_command_line_arguments()

    # Policy parameters
    policy_path = args.policy
    if policy_path is None:
        policy_path = f"{agent_dir}/policy/params.zip"
    if policy_path.endswith(".zip"):
        policy_path = policy_path[:-4]
    logging.info("Loading policy from %s.zip", policy_path)

    # Configuration
    config_path = f"{os.path.dirname(policy_path)}/operative_config.gin"
    logging.info("Loading policy configuration from %s", config_path)
    gin.parse_config_file(config_path)

    try:
        main(policy_path, args.training)
    except KeyboardInterrupt:
        logging.info("Caught a keyboard interrupt")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Wheel balancing using model predictive control with the ProxQP solver."""

import argparse
import os
import time
from time import perf_counter
from typing import Optional

import gin
import gymnasium as gym
import numpy as np
import proxsuite
import qpsolvers
from numpy.typing import NDArray
from qpmpc import MPCQP, Plan, solve_mpc
from qpmpc.systems import WheeledInvertedPendulum
from qpsolvers import solve_problem

import upkie.envs
from upkie.utils.clamp import clamp_and_warn
from upkie.utils.filters import low_pass_filter
from upkie.utils.raspi import configure_agent_process, on_raspi
from upkie.utils.spdlog import logging

upkie.envs.register()


def parse_command_line_arguments() -> argparse.Namespace:
    """
    Parse command line arguments.

    Returns:
        Command-line arguments.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--live-plot",
        help="Display a live plot of MPC trajectories",
        action="store_true",
        default=False,
    )
    return parser.parse_args()


@gin.configurable
class ProxQPWorkspace:
    def __init__(
        self, mpc_qp: MPCQP, update_preconditioner: bool, verbose: bool
    ):
        n_eq = 0
        n_in = mpc_qp.h.size // 2  # WheeledInvertedPendulum structure
        n = mpc_qp.P.shape[1]
        solver = proxsuite.proxqp.dense.QP(
            n,
            n_eq,
            n_in,
            dense_backend=proxsuite.proxqp.dense.DenseBackend.PrimalDualLDLT,
        )
        solver.settings.eps_abs = 1e-3
        solver.settings.eps_rel = 0.0
        solver.settings.verbose = verbose
        solver.settings.compute_timings = True
        solver.settings.primal_infeasibility_solving = True
        solver.init(
            H=mpc_qp.P,
            g=mpc_qp.q,
            C=mpc_qp.G[::2, :],  # WheeledInvertedPendulum structure
            l=-mpc_qp.h[1::2],  # WheeledInvertedPendulum structure
            u=mpc_qp.h[::2],  # WheeledInvertedPendulum structure
        )
        solver.solve()
        self.update_preconditioner = update_preconditioner
        self.solver = solver

    def solve(self, mpc_qp: MPCQP) -> qpsolvers.Solution:
        self.solver.update(
            g=mpc_qp.q,
            update_preconditioner=self.update_preconditioner,
        )
        self.solver.solve()
        qpsol = qpsolvers.Solution(mpc_qp.problem)
        qpsol.found = True
        qpsol.x = self.solver.results.x
        return qpsol


@gin.configurable
class PendularUpkie(WheeledInvertedPendulum):
    def __init__(
        self,
        length: float,
        max_ground_accel: float,
        nb_timesteps: int,
        sampling_period: float,
    ):
        super().__init__(
            length=length,
            max_ground_accel=max_ground_accel,
            nb_timesteps=nb_timesteps,
            sampling_period=sampling_period,
        )


@gin.configurable
def balance(
    env: gym.Env,
    nb_env_steps: int,
    rebuild_qp_every_time: bool,
    show_live_plot: bool,
    stage_input_cost_weight: float,
    stage_state_cost_weight: float,
    terminal_cost_weight: float,
    warm_start: bool,
):
    """!
    Run proportional balancer in gym environment with logging.

    @param env Gym environment to Upkie.
    @param nb_env_steps Number of environment steps to perform (zero to run
        indefinitely).
    @param rebuild_qp_every_time If set, rebuild all QP matrices at every
        iteration. Otherwise, only update vectors.
    @param show_live_plot Show a live plot.
    @param stage_input_cost_weight Weight for the stage input cost.
    @param stage_state_cost_weight Weight for the stage state cost.
    @param terminal_cost_weight Weight for the terminal cost.
    @param warm_start If set, use the warm-starting feature of ProxQP.
    """
    pendulum = PendularUpkie()
    mpc_problem = pendulum.build_mpc_problem(
        terminal_cost_weight=terminal_cost_weight,
        stage_state_cost_weight=stage_state_cost_weight,
        stage_input_cost_weight=stage_input_cost_weight,
    )
    mpc_problem.initial_state = np.zeros(4)
    mpc_qp = MPCQP(mpc_problem)
    proxqp = ProxQPWorkspace(mpc_qp)

    live_plot = None
    if show_live_plot and not on_raspi():
        from qpmpc.live_plots import (  # imports matplotlib
            WheeledInvertedPendulumPlot,
        )

        live_plot = WheeledInvertedPendulumPlot(pendulum, order="velocities")

    env.reset()  # connects to the spine
    commanded_velocity = 0.0
    action = np.zeros(env.action_space.shape)

    planning_times = np.empty((nb_env_steps,)) if nb_env_steps > 0 else None
    base_pitches = np.empty((nb_env_steps,)) if nb_env_steps > 0 else None
    step = 0
    while True:
        action[0] = commanded_velocity
        observation, _, terminated, truncated, info = env.step(action)
        if terminated or truncated:
            observation, info = env.reset()
            commanded_velocity = 0.0

        spine_observation = info["spine_observation"]
        floor_contact = spine_observation["floor_contact"]["contact"]

        # Unpack observation into initial MPC state
        (
            base_pitch,
            ground_position,
            base_angular_velocity,
            ground_velocity,
        ) = observation
        initial_state = np.array(
            [
                ground_position,
                base_pitch,
                ground_velocity,
                base_angular_velocity,
            ]
        )

        nx = WheeledInvertedPendulum.STATE_DIM
        target_states = np.zeros((pendulum.nb_timesteps + 1) * nx)
        mpc_problem.update_initial_state(initial_state)
        mpc_problem.update_goal_state(target_states[-nx:])
        mpc_problem.update_target_states(target_states[:-nx])

        t0 = perf_counter()
        if rebuild_qp_every_time:
            plan = solve_mpc(mpc_problem, solver="proxqp")
        else:
            mpc_qp.update_cost_vector(mpc_problem)
            if warm_start:
                qpsol = proxqp.solve(mpc_qp)
            else:
                qpsol = solve_problem(mpc_qp.problem, solver="proxqp")
            plan = Plan(mpc_problem, qpsol)
        if nb_env_steps > 0:
            base_pitches[step] = base_pitch
            planning_times[step] = perf_counter() - t0

        if not floor_contact:
            commanded_velocity = low_pass_filter(
                prev_output=commanded_velocity,
                cutoff_period=0.1,
                new_input=0.0,
                dt=env.dt,
            )
        elif plan.is_empty:
            logging.error("Solver found no solution to the MPC problem")
            logging.info("Continuing with previous action")
        else:  # plan was found
            pendulum.state = initial_state
            if live_plot is not None:
                t = time.time()
                live_plot.update(plan, t, initial_state, t)
            commanded_accel = plan.first_input
            commanded_velocity = clamp_and_warn(
                commanded_velocity + commanded_accel * env.dt / 2.0,
                lower=-1.0,
                upper=+1.0,
                label="commanded_velocity",
            )

        if nb_env_steps > 0:
            step += 1
            if step >= nb_env_steps:
                break

    report(mpc_problem, mpc_qp, planning_times)
    np.save("base_pitches.npy", base_pitches)
    np.save("planning_times.npy", planning_times)


def report(mpc_problem, mpc_qp, planning_times: Optional[NDArray[float]]):
    average_ms = 1e3 * np.average(planning_times)
    std_ms = 1e3 * np.std(planning_times)
    nb_env_steps = planning_times.size
    print("")
    print(f"{gin.operative_config_str()}")
    print(f"{mpc_problem.goal_state=}")
    print(f"{mpc_problem.nb_timesteps=}")
    print(f"{mpc_qp.P.shape=}")
    print(f"{mpc_qp.q.shape=}")
    print(f"{mpc_qp.G.shape=}")
    print(f"{mpc_qp.h.shape=}")
    print(f"{mpc_qp.Phi.shape=}")
    print(f"{mpc_qp.Psi.shape=}")
    print("")
    if planning_times is not None:
        print(
            "Planning time: "
            f"{average_ms:.2} ± {std_ms:.2} ms over {nb_env_steps} calls"
        )
        print("")


if __name__ == "__main__":
    if on_raspi():
        configure_agent_process()

    agent_dir = os.path.dirname(__file__)
    gin.parse_config_file(f"{agent_dir}/config.gin")
    args = parse_command_line_arguments()
    with gym.make("UpkieGroundVelocity-v3", frequency=200.0) as env:
        balance(env, show_live_plot=args.live_plot)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test NoisifyObservation wrapper."""

import unittest

import gymnasium
import numpy as np

from upkie.envs.wrappers.noisify_observation import NoisifyObservation
from upkie.envs.wrappers.tests.envs import ConstantObservationEnv


class NoisifyObservationTestCase(unittest.TestCase):
    def test_noise(self):
        env = ConstantObservationEnv(1.0)
        observation, _, _, _, _ = env.step(None)
        self.assertLess(abs(observation[0] - 1.0), 1e-10)

        noisy_env = NoisifyObservation(env, noise=np.array([0.42]))
        observation, _, _, _, _ = noisy_env.step(None)
        self.assertGreater(abs(observation[0] - 1.0), 1e-10)

    def test_check_env(self):
        try:
            from stable_baselines3.common.env_checker import check_env

            env = gymnasium.make("Acrobot-v1")
            noisy_env = NoisifyObservation(env, noise=np.full(6, 0.42))
            check_env(noisy_env)
        except ImportError:
            pass


if __name__ == "__main__":
    unittest.main()  # necessary for `bazel test`



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

import argparse
import datetime
import os
import random
import signal
import tempfile
from typing import Callable, List

import gin
import gymnasium
import numpy as np
import stable_baselines3
from envs import make_ppo_balancer_env
from rules_python.python.runfiles import runfiles
from settings import EnvSettings, PPOSettings, TrainingSettings
from stable_baselines3.common.callbacks import BaseCallback, CheckpointCallback
from stable_baselines3.common.logger import TensorBoardOutputFormat
from stable_baselines3.common.monitor import Monitor
from stable_baselines3.common.utils import set_random_seed
from stable_baselines3.common.vec_env import DummyVecEnv, SubprocVecEnv
from stable_baselines3.common.vec_env.base_vec_env import VecEnv
from torch import nn

import upkie.envs
from upkie.utils.spdlog import logging

upkie.envs.register()


def parse_command_line_arguments() -> argparse.Namespace:
    """
    Parse command line arguments.

    Returns:
        Command-line arguments.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--name",
        default="",
        type=str,
        help="name of the new policy to train",
    )
    parser.add_argument(
        "--nb-envs",
        default=1,
        type=int,
        help="number of parallel simulation processes to run",
    )
    parser.add_argument(
        "--show",
        default=False,
        action="store_true",
        help="show simulator during trajectory rollouts",
    )
    return parser.parse_args()


class InitRandomizationCallback(BaseCallback):
    def __init__(
        self,
        vec_env: VecEnv,
        key: str,
        max_value: float,
        start_timestep: int,
        end_timestep: int,
    ):
        super().__init__()
        self.end_timestep = end_timestep
        self.key = key
        self.max_value = max_value
        self.start_timestep = start_timestep
        self.vec_env = vec_env

    def _on_step(self) -> bool:
        progress: float = np.clip(
            (self.num_timesteps - self.start_timestep) / self.end_timestep,
            0.0,
            1.0,
        )
        cur_value = progress * self.max_value
        self.vec_env.env_method("update_init_rand", **{self.key: cur_value})
        self.logger.record(f"init_rand/{self.key}", cur_value)


class SummaryWriterCallback(BaseCallback):
    def __init__(self, vec_env: VecEnv, save_path: str):
        super().__init__()
        self.save_path = save_path
        self.vec_env = vec_env

    def _on_training_start(self):
        output_formats = self.logger.output_formats
        self.tb_formatter = next(
            formatter
            for formatter in output_formats
            if isinstance(formatter, TensorBoardOutputFormat)
        )

    def _on_step(self) -> bool:
        # We wait for the first call to log operative config so that parameters
        # for functions called by the environment are logged as well.
        if self.n_calls != 1:
            return
        self.tb_formatter.writer.add_text(
            "gin/operative_config",
            gin.operative_config_str(),
            global_step=None,
        )
        gin_path = f"{self.save_path}/operative_config.gin"
        with open(gin_path, "w") as fh:
            fh.write(gin.operative_config_str())
        logging.info(f"Saved gin configuration to {gin_path}")


def get_random_word():
    with open("/usr/share/dict/words") as fh:
        words = fh.read().splitlines()
    word_index = random.randint(0, len(words))
    while not words[word_index].isalnum():
        word_index = (word_index + 1) % len(words)
    return words[word_index]


def get_bullet_argv(shm_name: str, show: bool) -> List[str]:
    """!
    Get command-line arguments for the Bullet spine.

    @param shm_name Name of the shared-memory file.
    @param show If true, show simulator GUI.
    @returns Command-line arguments.
    """
    env_settings = EnvSettings()
    agent_frequency = env_settings.agent_frequency
    spine_frequency = env_settings.spine_frequency
    assert spine_frequency % agent_frequency == 0
    nb_substeps = spine_frequency / agent_frequency
    bullet_argv = []
    bullet_argv.extend(["--shm-name", shm_name])
    bullet_argv.extend(["--nb-substeps", str(nb_substeps)])
    bullet_argv.extend(["--spine-frequency", str(spine_frequency)])
    if show:
        bullet_argv.append("--show")
    return bullet_argv


def init_env(
    max_episode_duration: float,
    show: bool,
    spine_path: str,
):
    """!
    Get an environment initialization function for a set of parameters.

    @param max_episode_duration Maximum duration of an episode, in seconds.
    @param show If true, show simulator GUI.
    @param spine_path Path to the Bullet spine binary.
    """
    env_settings = EnvSettings()
    seed = random.randint(0, 1_000_000)

    def _init():
        shm_name = f"/{get_random_word()}"
        pid = os.fork()
        if pid == 0:  # child process: spine
            argv = get_bullet_argv(shm_name, show=show)
            os.execvp(spine_path, ["bullet"] + argv)
            return

        # parent process: trainer
        agent_frequency = env_settings.agent_frequency
        velocity_env = gymnasium.make(
            env_settings.env_id,
            max_episode_steps=int(max_episode_duration * agent_frequency),
            frequency=agent_frequency,
            regulate_frequency=False,
            reward_weights=upkie.envs.UpkieGroundVelocity.RewardWeights(
                **env_settings.reward_weights
            ),
            shm_name=shm_name,
            spine_config=env_settings.spine_config,
            max_ground_velocity=env_settings.max_ground_velocity,
        )
        velocity_env.reset(seed=seed)
        velocity_env._prepatch_close = velocity_env.close

        def close_monkeypatch():
            logging.info(f"Terminating spine {shm_name} with {pid=}...")
            os.kill(pid, signal.SIGINT)  # interrupt spine child process
            os.waitpid(pid, 0)  # wait for spine to terminate
            velocity_env._prepatch_close()

        velocity_env.close = close_monkeypatch
        env = make_ppo_balancer_env(velocity_env, env_settings, training=True)
        return Monitor(env)

    set_random_seed(seed)
    return _init


def find_save_path(training_dir: str, policy_name: str):
    def path_for_iter(nb_iter: int):
        return f"{training_dir}/{policy_name}_{nb_iter}"

    nb_iter = 1
    while os.path.exists(path_for_iter(nb_iter)):
        nb_iter += 1
    return path_for_iter(nb_iter)


def affine_schedule(y_0: float, y_1: float) -> Callable[[float], float]:
    """!
    Affine schedule as a function over the [0, 1] interval.

    @param y_0 Function value at zero.
    @param y_1 Function value at one.
    @return Corresponding affine function.
    """
    diff = y_1 - y_0

    def schedule(x: float) -> float:
        """!
        Compute the current learning rate from remaining progress.

        @param x Progress decreasing from 1 (beginning) to 0.
        @return Corresponding learning rate>
        """
        return y_0 + x * diff

    return schedule


def train_policy(
    policy_name: str,
    training_dir: str,
    nb_envs: int,
    show: bool,
) -> None:
    """!
    Train a new policy and save it to a directory.

    @param policy_name Name of the trained policy.
    @param training_dir Directory for logging and saving policies.
    @param nb_envs Number of environments, each running in a separate process.
    @param show Whether to show the simulation GUI.
    """
    if policy_name == "":
        policy_name = get_random_word()
    save_path = find_save_path(training_dir, policy_name)
    logging.info('New policy name is "%s"', policy_name)
    logging.info("Training data will be logged to %s", save_path)

    training = TrainingSettings()
    deez_runfiles = runfiles.Create()
    spine_path = os.path.join(
        agent_dir,
        deez_runfiles.Rlocation("upkie/spines/bullet_spine"),
    )

    vec_env = (
        SubprocVecEnv(
            [
                init_env(
                    max_episode_duration=training.max_episode_duration,
                    show=show,
                    spine_path=spine_path,
                )
                for i in range(nb_envs)
            ],
            start_method="fork",
        )
        if nb_envs > 1
        else DummyVecEnv(
            [
                init_env(
                    max_episode_duration=training.max_episode_duration,
                    show=show,
                    spine_path=spine_path,
                )
            ]
        )
    )

    env_settings = EnvSettings()
    dt = 1.0 / env_settings.agent_frequency
    gamma = 1.0 - dt / training.return_horizon
    logging.info(
        "Discount factor gamma=%f for a return horizon of %f s",
        gamma,
        training.return_horizon,
    )

    ppo_settings = PPOSettings()
    policy = stable_baselines3.PPO(
        "MlpPolicy",
        vec_env,
        learning_rate=affine_schedule(
            y_1=ppo_settings.learning_rate,  # progress_remaining=1.0
            y_0=ppo_settings.learning_rate / 3,  # progress_remaining=0.0
        ),
        n_steps=ppo_settings.n_steps,
        batch_size=ppo_settings.batch_size,
        n_epochs=ppo_settings.n_epochs,
        gamma=gamma,
        gae_lambda=ppo_settings.gae_lambda,
        clip_range=ppo_settings.clip_range,
        clip_range_vf=ppo_settings.clip_range_vf,
        normalize_advantage=ppo_settings.normalize_advantage,
        ent_coef=ppo_settings.ent_coef,
        vf_coef=ppo_settings.vf_coef,
        max_grad_norm=ppo_settings.max_grad_norm,
        use_sde=ppo_settings.use_sde,
        sde_sample_freq=ppo_settings.sde_sample_freq,
        target_kl=ppo_settings.target_kl,
        tensorboard_log=training_dir,
        policy_kwargs={
            "activation_fn": nn.Tanh,
            "net_arch": {
                "pi": ppo_settings.net_arch_pi,
                "vf": ppo_settings.net_arch_vf,
            },
        },
        verbose=1,
    )

    try:
        policy.learn(
            total_timesteps=training.total_timesteps,
            callback=[
                CheckpointCallback(
                    save_freq=max(210_000 // nb_envs, 1_000),
                    save_path=save_path,
                    name_prefix="checkpoint",
                ),
                SummaryWriterCallback(vec_env, save_path),
                InitRandomizationCallback(
                    vec_env,
                    "pitch",
                    training.init_rand["pitch"],
                    start_timestep=0,
                    end_timestep=1e5,
                ),
                InitRandomizationCallback(
                    vec_env,
                    "v_x",
                    training.init_rand["v_x"],
                    start_timestep=0,
                    end_timestep=1e5,
                ),
                InitRandomizationCallback(
                    vec_env,
                    "omega_y",
                    training.init_rand["omega_y"],
                    start_timestep=0,
                    end_timestep=1e5,
                ),
            ],
            tb_log_name=policy_name,
        )
    except KeyboardInterrupt:
        logging.info("Training interrupted.")

    # Save policy no matter what!
    policy.save(f"{save_path}/final.zip")
    policy.env.close()


if __name__ == "__main__":
    args = parse_command_line_arguments()
    agent_dir = os.path.dirname(__file__)
    gin.parse_config_file(f"{agent_dir}/settings.gin")

    training_path = os.environ.get(
        "UPKIE_TRAINING_PATH", tempfile.gettempdir()
    )
    date = datetime.datetime.now().strftime("%Y-%m-%d")
    training_dir = f"{training_path}/{date}"
    logging.info("Logging training data in %s", training_dir)
    logging.info(
        "To track in TensorBoard, run "
        f"`tensorboard --logdir {training_dir}`"
    )
    train_policy(
        args.name,
        training_dir,
        nb_envs=args.nb_envs,
        show=args.show,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test NoisifyAction wrapper."""

import unittest

import gymnasium
import numpy as np

from upkie.envs.wrappers.noisify_action import NoisifyAction
from upkie.envs.wrappers.tests.envs import ActionObserverEnv


class NoisifyActionTestCase(unittest.TestCase):
    def test_noise(self):
        noisy_env = NoisifyAction(ActionObserverEnv(), noise=np.array([0.42]))
        action = np.array([1.0])
        inner_action, _, _, _, _ = noisy_env.step(action)
        self.assertGreater(np.abs(inner_action - action), 1e-10)

    def test_check_env(self):
        try:
            from stable_baselines3.common.env_checker import check_env

            env = gymnasium.make("Pendulum-v1")
            noisy_env = NoisifyAction(env, noise=np.full(1, 0.42))
            check_env(noisy_env)
        except ImportError:
            pass


if __name__ == "__main__":
    unittest.main()  # necessary for `bazel test`



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""
Test base pitch calculation.
"""

import unittest

import numpy as np

from upkie.observers.base_pitch import (
    compute_base_pitch_from_imu,
    compute_pitch_frame_in_parent,
)


class TestBasePitch(unittest.TestCase):
    def test_zero_pitch(self, phi=0.42):
        """
        Check that a pure rotation around the z-axis yields no pitch.
        """
        orientation_imu_in_world = np.array(
            [
                [np.cos(phi), -np.sin(phi), 0.0],
                [np.sin(phi), np.cos(phi), 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        imu_pitch = compute_pitch_frame_in_parent(orientation_imu_in_world)
        self.assertEqual(imu_pitch, 0.0)

    def test_close_to_zero(self, theta=1e-3):
        """
        Check that a pure rotation around the y-axis yields the correct pitch.
        """
        orientation_imu_in_world = np.array(
            [
                [np.cos(theta), 0.0, np.sin(theta)],
                [0.0, 1.0, 0.0],
                [-np.sin(theta), 0.0, np.cos(theta)],
            ]
        )
        imu_pitch = compute_pitch_frame_in_parent(orientation_imu_in_world)
        self.assertTrue(np.isclose(imu_pitch, theta))

    def test_orientation_not_neatly_normalized(self, theta=1e-3):
        """
        As an extra, the function can handle a partially-normalized input. We
        are going the extra mile here, this is not strictly necessary.
        """
        orientation_imu_in_world = np.array(
            [
                [np.cos(theta), 0.0, np.sin(theta)],
                [0.0, 1.0, 0.0],
                [-np.sin(theta), 0.0, np.cos(theta)],
            ]
        )
        orientation_imu_in_world[:, 0] *= 1.0 - 1e-2
        imu_pitch = compute_pitch_frame_in_parent(orientation_imu_in_world)
        self.assertTrue(np.isclose(imu_pitch, theta))

    def test_balancing_plane_lateral_in_imu_frame(self):
        quat_imu_in_ars = np.array(
            [
                0.008472769239730098,
                -0.9953038144146671,
                -0.09639792825405252,
                -0.002443076206500708,
            ]
        )
        rotation_base_to_imu = np.array(
            [[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]
        )
        pitch_base_in_world = compute_base_pitch_from_imu(
            quat_imu_in_ars, rotation_base_to_imu
        )
        self.assertAlmostEqual(pitch_base_in_world, -0.016, places=3)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Tests for UpkieServos environment."""

import unittest

import numpy as np
import posix_ipc

from upkie.envs import UpkieServos
from upkie.envs.tests.mock_spine import MockSpine
from upkie.utils.exceptions import ActionError


class TestUpkieServos(unittest.TestCase):
    def setUp(self):
        shm_name = "/vroum"
        shared_memory = posix_ipc.SharedMemory(
            shm_name, posix_ipc.O_RDWR | posix_ipc.O_CREAT, size=42
        )
        self.env = UpkieServos(
            fall_pitch=1.0,
            frequency=100.0,
            shm_name=shm_name,
        )
        shared_memory.close_fd()
        self.env._spine = MockSpine()

    def test_reset(self):
        observation, info = self.env.reset()
        spine_observation = info["spine_observation"]
        self.assertAlmostEqual(
            observation["wheel_odometry"]["position"],
            spine_observation["wheel_odometry"]["position"],
        )
        self.assertGreaterEqual(spine_observation["number"], 1)

        imu = observation["imu"]
        self.assertIsInstance(imu["angular_velocity"], np.ndarray)
        self.assertIsInstance(imu["linear_acceleration"], np.ndarray)
        self.assertIsInstance(imu["orientation"], np.ndarray)
        self.assertEqual(imu["angular_velocity"].shape, (3,))
        self.assertEqual(imu["linear_acceleration"].shape, (3,))
        self.assertEqual(imu["orientation"].shape, (4,))

    def test_reward(self):
        self.env.reset()
        action = {
            servo: {
                "position": np.nan,
                "velocity": 0.0,
            }
            for servo in self.env.JOINT_NAMES
        }
        _, reward, _, _, _ = self.env.step(action)
        self.assertAlmostEqual(reward, 1.0)  # survival reward

    def test_action_needs_position(self):
        observation, info = self.env.reset()
        action = {servo: {"velocity": 0.0} for servo in self.env.JOINT_NAMES}
        with self.assertRaises(ActionError):
            self.env.step(action)

    def test_action_clamping(self):
        action = {
            servo: {
                "position": np.nan,
                "velocity": 0.0,
                "feedforward_torque": 0.0,
            }
            for servo in self.env.JOINT_NAMES
        }
        not_wheel = "left_hip"  # wheels don't have position limits
        self.env.reset()

        action[not_wheel]["position"] = np.nan
        self.env.step(action)
        self.assertTrue(
            np.isnan(self.env._spine.action["servo"][not_wheel]["position"])
        )

        action[not_wheel]["position"] = 0.5
        self.env.step(action)
        self.assertAlmostEqual(
            self.env._spine.action["servo"][not_wheel]["position"],
            0.5,
            places=5,
        )

        action[not_wheel]["position"] = 5e5
        self.env.step(action)
        self.assertAlmostEqual(
            self.env._spine.action["servo"][not_wheel]["position"],
            float(self.env.action_space[not_wheel]["position"].high),
            places=5,
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

"""Keep the robot up using its wheels."""

from typing import Tuple

import gin
import numpy as np
from numpy.typing import NDArray

from upkie.observers.base_pitch import compute_base_pitch_from_imu
from upkie.utils.clamp import clamp, clamp_abs
from upkie.utils.filters import abs_bounded_derivative_filter, low_pass_filter


@gin.configurable
class WheelController:

    """
    Balancing by proportional-derivative feedback of the body pitch error to
    wheel accelerations:

        body pitch error --(PD)--> wheel accelerations

    Attributes:
        air_return_period: Cutoff period for resetting integrators while the
            robot is in the air, in [s].
        error: Two-dimensional vector of ground position and base pitch errors.
        gains: velocity controller gains.
        ground_velocity: Sagittal velocity in [m] / [s].
        integral_error_velocity: Integral term contributing to the sagittal
            velocity, in [m] / [s].
        max_ground_velocity: Maximum commanded ground velocity no matter what,
            in [m] / [s].
        max_integral_error_velocity: Maximum integral error velocity, in [m] /
            [s].
        max_target_accel: Maximum acceleration for the ground target, in
            [m] / [s]². Does not affect the commanded ground velocity.
        max_target_distance: Maximum distance from the current ground position
            to the target, in [m].
        max_target_velocity: Maximum velocity for the ground target, in
            [m] / [s]. Indirectly affects the commanded ground velocity.
        pitch: Current IMU pitch angle in [rad].
        target_ground_position: Target ground sagittal position in [m].
        target_ground_velocity: Target ground sagittal velocity in [m] / [s].
        target_yaw_position: Target yaw position in [rad].
        target_yaw_velocity: Target yaw velocity in [rad] / [s].
        turning_deadband: Joystick axis value between 0.0 and 1.0 below which
            legs stiffen but the turning motion doesn't start.
        turning_probability: Probability that the user wants to turn based on
            the joystick axis value.
        turning_decision_time: Minimum duration in [s] for the turning
            probability to switch from zero to one and converesly.
        wheel_radius: Wheel radius in [m].
    """

    @gin.configurable
    class Gains:
        pitch_damping: float
        pitch_stiffness: float
        position_damping: float
        position_stiffness: float

        def __init__(
            self,
            pitch_damping: float,
            pitch_stiffness: float,
            position_damping: float,
            position_stiffness: float,
        ):
            self.pitch_damping = pitch_damping
            self.pitch_stiffness = pitch_stiffness
            self.position_damping = position_damping
            self.position_stiffness = position_stiffness

        def set(
            self,
            pitch_damping: float,
            pitch_stiffness: float,
            position_damping: float,
            position_stiffness: float,
        ) -> None:
            """!
            Set gains in one function call.

            @param pitch_damping: Pitch error (normalized) damping gain.
                Corresponds to the proportional term of the velocity PI
                controller, equivalent to the derivative term of the
                acceleration PD controller.
            @param pitch_stiffness: Pitch error (normalized) stiffness gain.
                Corresponds to the integral term of the velocity PI controller,
                equivalent to the proportional term of the acceleration PD
                controller.
            @param position_damping: Position error (normalized) damping gain.
                Corresponds to the proportional term of the velocity PI
                controller, equivalent to the derivative term of the
                acceleration PD controller.
            @param position_stiffness: Position error (normalized) stiffness
                gain. Corresponds to the integral term of the velocity PI
                controller, equivalent to the proportional term of the
                acceleration PD controller.
            """
            self.pitch_damping = pitch_damping
            self.pitch_stiffness = pitch_stiffness
            self.position_damping = position_damping
            self.position_stiffness = position_stiffness

        def __repr__(self):
            return (
                "WheelController.Gains("
                f"pitch_stiffness={self.pitch_stiffness}, "
                f"pitch_damping={self.pitch_damping}, "
                f"position_stiffness={self.position_stiffness}, "
                f"position_damping={self.position_damping})"
            )

    air_return_period: float
    error: NDArray[float]
    gains: Gains
    ground_velocity: float
    integral_error_velocity: float
    max_ground_velocity: float
    max_integral_error_velocity: float
    max_target_accel: float
    max_target_distance: float
    max_target_velocity: float
    pitch: float
    target_ground_position: float
    target_ground_velocity: float
    target_yaw_position: float
    target_yaw_velocity: float
    turning_deadband: float
    turning_decision_time: float
    turning_probability: float
    wheel_radius: float

    def __init__(
        self,
        air_return_period: float,
        fall_pitch: float,
        max_ground_velocity: float,
        max_integral_error_velocity: float,
        max_target_accel: float,
        max_target_distance: float,
        max_target_velocity: float,
        max_yaw_accel: float,
        max_yaw_velocity: float,
        turning_deadband: float,
        turning_decision_time: float,
        wheel_radius: float,
    ):
        """!
        Initialize balancer.

        @param air_return_period Cutoff period for resetting integrators while
            the robot is in the air, in [s].
        @param fall_pitch Fall pitch angle, in radians.
        @param max_ground_velocity Maximum commanded ground velocity no matter
            what, in [m] / [s].
        @param max_integral_error_velocity Maximum integral error velocity, in
            [m] / [s].
        @param max_target_accel Maximum acceleration for the ground target, in
            [m] / [s]². This bound does not affect the commanded ground
            velocity.
        @param max_target_distance Maximum distance from the current ground
            position to the target, in [m].
        @param max_target_velocity Maximum velocity for the ground target, in
            [m] / [s]. This bound indirectly affects the commanded ground
            velocity.
        @param max_yaw_accel Maximum yaw angular acceleration in [rad] / [s]².
        @param max_yaw_velocity Maximum yaw angular velocity in [rad] / [s].
        @param turning_deadband Joystick axis value between 0.0 and 1.0 below
            which legs stiffen but the turning motion doesn't start.
        @param turning_decision_time Minimum duration in [s] for the turning
            probability to switch from zero to one and converesly.
        @param wheel_radius Wheel radius in [m].
        """
        assert 0.0 <= turning_deadband <= 1.0
        self.air_return_period = air_return_period
        self.error = np.zeros(2)
        self.fall_pitch = fall_pitch
        self.gains = WheelController.Gains()  # type: ignore
        self.ground_velocity = 0.0
        self.integral_error_velocity = 0.0
        self.max_ground_velocity = max_ground_velocity
        self.max_integral_error_velocity = max_integral_error_velocity
        self.max_target_accel = max_target_accel
        self.max_target_distance = max_target_distance
        self.max_target_velocity = max_target_velocity
        self.max_yaw_accel = max_yaw_accel
        self.max_yaw_velocity = max_yaw_velocity
        self.pitch = 0.0
        self.target_ground_position = 0.0
        self.target_ground_velocity = 0.0
        self.target_yaw_position = 0.0
        self.target_yaw_velocity = 0.0
        self.turning_deadband = turning_deadband
        self.turning_decision_time = turning_decision_time
        self.turning_probability = 0.0
        self.wheel_radius = wheel_radius

    def update_target_ground_velocity(
        self, observation: dict, dt: float
    ) -> None:
        """!
        Update target ground velocity from joystick input.

        @param observation Latest observation.
        @param dt Time in [s] until next cycle.

        @note The target ground velocity is commanded by both the left axis and
        right trigger of the joystick. When the right trigger is unpressed, the
        commanded velocity is set from the left axis, interpolating from 0 to
        50% of its maximum configured value. Pressing the right trigger
        increases it further up to 100% of the configured value.
        """
        try:
            axis_value = observation["joystick"]["left_axis"][1]
            trigger_value = observation["joystick"]["right_trigger"]  # -1 to 1
            boost_value = clamp_abs(0.5 * (trigger_value + 1.0), 1.0)  # 0 to 1
            max_velocity = 0.5 * (1.0 + boost_value) * self.max_target_velocity
            unfiltered_velocity = -max_velocity * axis_value
        except KeyError:
            unfiltered_velocity = 0.0
        self.target_ground_velocity = abs_bounded_derivative_filter(
            self.target_ground_velocity,
            unfiltered_velocity,
            dt,
            self.max_target_velocity,
            self.max_target_accel,
        )

    def update_target_yaw_velocity(self, observation: dict, dt: float) -> None:
        """!
        Update target yaw velocity from joystick input.

        @param observation Latest observation.
        @param dt Time in [s] until next cycle.
        """
        try:
            joystick_value = observation["joystick"]["right_axis"][0]
        except KeyError:
            joystick_value = 0.0
        joystick_abs = abs(joystick_value)
        joystick_sign = np.sign(joystick_value)

        turning_intent = joystick_abs / self.turning_deadband
        self.turning_probability = abs_bounded_derivative_filter(
            self.turning_probability,
            turning_intent,  # might be > 1.0
            dt,
            max_output=1.0,  # output is <= 1.0
            max_derivative=1.0 / self.turning_decision_time,
        )

        velocity_ratio = (joystick_abs - self.turning_deadband) / (
            1.0 - self.turning_deadband
        )
        velocity_ratio = max(0.0, velocity_ratio)
        velocity = self.max_yaw_velocity * joystick_sign * velocity_ratio
        turn_hasnt_started = abs(self.target_yaw_velocity) < 0.01
        turn_not_sure_yet = self.turning_probability < 0.99
        if turn_hasnt_started and turn_not_sure_yet:
            velocity = 0.0
        self.target_yaw_velocity = abs_bounded_derivative_filter(
            self.target_yaw_velocity,
            velocity,
            dt,
            self.max_yaw_velocity,
            self.max_yaw_accel,
        )
        if abs(self.target_yaw_velocity) > 0.01:  # still turning
            self.turning_probability = 1.0

    def process_joystick_buttons(self, observation: dict) -> None:
        ground_position = observation["wheel_odometry"]["position"]
        try:
            if observation["joystick"]["cross_button"]:
                # When the user presses the reset button, we assume there is no
                # contact for sure (or we are in a situation where spinning the
                # wheels is dangerous?) and thus perform a hard rather than
                # soft reset of both integrators.
                self.integral_error_velocity = 0.0  # [m] / [s]
                self.target_ground_position = ground_position
        except KeyError:
            pass

    def cycle(self, observation: dict, dt: float) -> None:
        """!
        Compute a new ground velocity.

        @param observation Latest observation.
        @param dt Time in [s] until next cycle.
        """
        self.process_joystick_buttons(observation)
        self.update_target_ground_velocity(observation, dt)
        self.update_target_yaw_velocity(observation, dt)

        pitch = compute_base_pitch_from_imu(observation["imu"]["orientation"])
        self.pitch = pitch
        if abs(pitch) > self.fall_pitch:
            self.integral_error_velocity = 0.0  # [m] / [s]
            self.ground_velocity = 0.0  # [m] / [s]
            return

        ground_position = observation["wheel_odometry"]["position"]
        floor_contact = observation["floor_contact"]["contact"]

        target_pitch: float = 0.0  # [rad]
        error = np.array(
            [
                self.target_ground_position - ground_position,
                target_pitch - pitch,
            ]
        )
        self.error = error

        if not floor_contact:
            self.integral_error_velocity = low_pass_filter(
                self.integral_error_velocity, self.air_return_period, 0.0, dt
            )
            # We don't reset self.target_ground_velocity: either takeoff
            # detection is a false positive and we should resume close to the
            # pre-takeoff state, or the robot is really in the air and the user
            # should stop smashing the joystick like a bittern ;p
            self.target_ground_position = low_pass_filter(
                self.target_ground_position,
                self.air_return_period,
                ground_position,
                dt,
            )
        else:  # floor_contact:
            ki = np.array(
                [
                    self.gains.position_stiffness,
                    self.gains.pitch_stiffness,
                ]
            )
            self.integral_error_velocity += ki.dot(error) * dt
            self.integral_error_velocity = clamp_abs(
                self.integral_error_velocity, self.max_integral_error_velocity
            )
            self.target_ground_position += self.target_ground_velocity * dt
            self.target_ground_position = clamp(
                self.target_ground_position,
                ground_position - self.max_target_distance,
                ground_position + self.max_target_distance,
            )

        kp = np.array(
            [
                self.gains.position_damping,
                self.gains.pitch_damping,
            ]
        )

        # Non-minimum phase trick: as per control theory's book, the proper
        # feedforward velocity should be ``+self.target_ground_velocity``.
        # However, it is with resolute purpose that it sends
        # ``-self.target_ground_velocity`` instead!
        #
        # Try both on the robot, you will see the difference :)
        #
        # This hack is not purely out of "esprit de contradiction". Changing
        # velocity is a non-minimum phase behavior (to accelerate forward, the
        # ZMP of the LIPM needs to move backward at first, then forward), and
        # our feedback can't realize that (it only takes care of balancing
        # around a stationary velocity).
        #
        # What's left? Our integrator! If we send the opposite of the target
        # velocity (or only a fraction of it, although 100% seems to do a good
        # job), Upkie will immediately start executing the desired non-minimum
        # phase behavior. The error will then grow and the integrator catch up
        # so that ``upkie_trick_velocity - self.integral_error_velocity``
        # converges to its proper steady state value (the same value ``0 -
        # self.integral_error_velocity`` would have converged to if we had no
        # feedforward).
        #
        # Unconvinced? Try it on the robot. You will feel Upkie's trick ;)
        #
        upkie_trick_velocity = -self.target_ground_velocity

        self.ground_velocity = (
            upkie_trick_velocity - kp.dot(error) - self.integral_error_velocity
        )
        self.ground_velocity = clamp_abs(
            self.ground_velocity, self.max_ground_velocity
        )

    def get_wheel_velocities(
        self,
        position_right_in_left: NDArray[float],
    ) -> Tuple[float, float]:
        """!
        Get left and right wheel velocities.

        @param position_right_in_left: Translation from the left contact frame
            to the right contact frame, expressed in the left contact frame.
            Equivalently, linear coordinates of the pose of the right contact
            frame with respect to the left contact frame.

        @return Tuple with `left_wheel_velocity` (left wheel velocity in rad/s)
            and `right_wheel_velocity` (right wheel velocity in rad/s).

        @note For now we assume that the two wheels are parallel to the ground,
        so that the rotation from one frame to the other is the identity.
        """
        # Sagittal translation
        left_wheel_velocity: float = +self.ground_velocity / self.wheel_radius
        right_wheel_velocity: float = -self.ground_velocity / self.wheel_radius

        # Yaw rotation
        contact_radius = 0.5 * np.linalg.norm(position_right_in_left)
        yaw_to_wheel = contact_radius / self.wheel_radius
        left_wheel_velocity += yaw_to_wheel * self.target_yaw_velocity
        right_wheel_velocity += yaw_to_wheel * self.target_yaw_velocity

        return left_wheel_velocity, right_wheel_velocity

    def log(self) -> dict:
        """
        Log internal state to a dictionary.

        Returns:
            Log data as a dictionary.
        """
        return {
            "error": self.error,
            "gains": self.gains.__dict__,
            "ground_velocity": self.ground_velocity,
            "integral_error_velocity": self.integral_error_velocity,
            "pitch": self.pitch,
            "target_ground_position": self.target_ground_position,
            "target_ground_velocity": self.target_ground_velocity,
            "target_yaw_velocity": self.target_yaw_velocity,
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria
#
# This file incorporates work covered by the following copyright and
# permission notice:
#
#     setup-system.py from github.com:mjbots/quad
#     Copyright 2018-2020 Josh Pieper
#     SPDX-License-Identifier: Apache-2.0


"""Configure all servos connected to an Upkie wheeled biped."""

import datetime
import os
import subprocess
import sys
import time
from typing import Union

ORIG_SUFFIX = time.strftime(".orig-%Y%m%d-%H%M%S")

logging_depth = 0


def log_message(message: str, indent: int = 0) -> None:
    """Log a single message.

    Args:
        message: Message to log.
        indent: Indentation level.
    """
    logging_indent = " | " * (logging_depth + indent)
    now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    print(f"{now}: {logging_indent}{message}")


def log_method(func):
    """Decorator to log function calls."""

    def wrapped_function(*args, **kwargs):
        global logging_depth
        try:
            logging_depth += 1
            args_repr = (
                f"'{args[0]}'" + (", ..." if len(args) > 1 else "")
                if args
                else ""
            )
            log_message(f"{func.__name__}({args_repr})", indent=-1)
            result = func(*args, **kwargs)
            return result
        finally:
            logging_depth -= 1

    return wrapped_function


@log_method
def run(*args, **kwargs):
    """Run a subprocess."""
    subprocess.check_call(*args, shell=True, **kwargs)


def configure_servo(id: int, param: str, value: Union[float, str]):
    """Configure a moteus controller.

    Args:
        id: Identifier of the servo.
        param: Parameter to configure.
        value: Value corresponding to the parameter.
    """
    moteus_command = f"conf set {param} {value}"
    shell_command = (
        f'echo "{moteus_command}" | '
        f'sudo moteus_tool --pi3hat-cfg "1=1,2,3;2=4,5,6" -t {id} -c'
    )
    run(shell_command)


def write_configuration(id: int):
    """Write configuration to a moteus controller.

    Args:
        id: Identifier of the servo that should write its configuration.
    """
    shell_command = (
        f'echo "conf write" | '
        f'sudo moteus_tool --pi3hat-cfg "1=1,2,3;2=4,5,6" -t {id} -c'
    )
    run(shell_command)


if __name__ == "__main__":
    if os.geteuid() != 0:
        args = ["sudo", "-E", sys.executable] + sys.argv + [os.environ]
        os.execlpe("sudo", *args)

    # TODO(scaron): upkie layout for Python as well
    left_hip = 1
    left_knee = 2
    left_wheel = 3
    right_hip = 4
    right_knee = 5
    right_wheel = 6
    all_servos = (1, 2, 3, 4, 5, 6)

    for hip in (left_hip, right_hip):
        configure_servo(hip, "servopos.position_max", 0.2)
        configure_servo(hip, "servopos.position_min", -0.2)
        configure_servo(hip, "servo.max_velocity", 2.0)

    for knee in (left_knee, right_knee):
        configure_servo(knee, "servopos.position_max", 0.4)
        configure_servo(knee, "servopos.position_min", -0.4)
        configure_servo(knee, "servo.max_velocity", 2.0)

    for wheel in (left_wheel, right_wheel):
        configure_servo(wheel, "servopos.position_max", "NaN")
        configure_servo(wheel, "servopos.position_min", "NaN")
        configure_servo(wheel, "servo.max_velocity", 8.0)

    # https://github.com/mjbots/moteus/blob/38d688a933ce1584ee09f2628b5849d5e758ac21/docs/reference.md#servodefault_velocity_limit--servodefault_accel_limit
    for servo in all_servos:
        configure_servo(servo, "servo.default_velocity_limit", "NaN")
        configure_servo(servo, "servo.default_accel_limit", "NaN")

    for servo in all_servos:
        write_configuration(servo)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
#
# This file incorporates work covered by the following copyright and permission
# notice:
#
#     Copyright 2012-2016 Robot Locomotion Group @ CSAIL
#     License: BSD-3-Clause (see licenses/LICENSE-drake)

"""Linting with clang-format."""

import glob
import os
import platform
import shutil
import subprocess
import sys


def _is_cxx(filename):
    """Returns True only if filename is reasonable to clang-format."""
    root, ext = os.path.splitext(filename)
    _, penultimate_ext = os.path.splitext(root)

    # Don't clang-format CUDA files.
    if penultimate_ext == "cu":
        return False

    # Per https://bazel.build/versions/master/docs/be/c-cpp.html#cc_library
    return ext in [
        ".c",
        ".cc",
        ".cpp",
        ".cxx",
        ".c++",
        ".C",
        ".h",
        ".hh",
        ".hpp",
        ".hxx",
        ".inc",
    ]


def get_clang_format_path():
    """Get path to clang-format."""
    if platform.system() == "Darwin":
        candidates = glob.glob(
            "/opt/homebrew/*/clang-format/**/bin/clang-format",
            recursive=True,
        ) + ["/usr/local/opt/clang-format/bin/clang-format"]
        if len(candidates) < 1:
            raise RuntimeError("clang-format not found in homebrew paths")
        elif len(candidates) > 1:
            print(
                f"WARNING: multiple versions of clang-format: {candidates},"
                f"selecting {candidates[0]}"
            )
        path = candidates[0]
    else:  # platform.system() == "Linux"
        path = shutil.which("clang-format")
    if os.path.isfile(path):
        return path
    raise RuntimeError(f"clang-format not found at {path}")


def _check_clang_format_idempotence(filename):
    clang_format = get_clang_format_path()
    formatter = subprocess.Popen(
        [clang_format, "-style=file", filename], stdout=subprocess.PIPE
    )
    differ = subprocess.Popen(
        ["/usr/bin/diff", "-u", "-", filename],
        stdin=formatter.stdout,
        stdout=subprocess.PIPE,
    )
    changes = differ.communicate()[0]
    if not changes:
        return 0
    print("ERROR: {} needs clang-format".format(filename))
    print(f"You can fix it by `{clang_format} -style=file -i {filename}`")
    return 1


def main():
    """Checks that clang-format is idempotent.

    Checks are performed on each path specified as a command-line argument. We
    exit with code 1 if any of the paths are invalid or clang-format suggests
    any edits. Otherwise, we exit with code 0.
    """
    total_errors = 0
    for filename in sys.argv[1:]:
        if not _is_cxx(filename):
            print("clang_format_lint.py: Skipping " + filename)
            continue
        print("clang_format_lint.py: Linting " + filename)
        total_errors += _check_clang_format_idempotence(filename)
    if total_errors == 0:
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Test AddActionToObservation wrapper."""

import unittest

import gymnasium
import numpy as np

from upkie.envs.wrappers.add_action_to_observation import (
    AddActionToObservation,
)
from upkie.envs.wrappers.tests.envs import ConstantObservationEnv


class AddActionToObservationTestCase(unittest.TestCase):
    def test_wrapper(self):
        env = ConstantObservationEnv(42.0)
        wrapped_env = AddActionToObservation(env)
        action = np.array([1.0])
        observation, _, _, _, _ = wrapped_env.step(action)
        self.assertAlmostEqual(observation[0], 42.0)
        self.assertAlmostEqual(observation[1], 1.0)

    def test_check_env(self):
        try:
            from stable_baselines3.common.env_checker import check_env

            env = gymnasium.make("Pendulum-v1")
            wrapped_env = AddActionToObservation(env)
            check_env(wrapped_env)
        except ImportError:
            pass


if __name__ == "__main__":
    unittest.main()  # necessary for `bazel test`



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from dataclasses import dataclass
from typing import List, Optional, Tuple

import gin


@gin.configurable
@dataclass
class EnvSettings:
    """!
    Environment settings.
    """

    accel_penalty: float
    action_lpf: Tuple[float, float]
    action_noise: List[float]
    agent_frequency: int
    env_id: str
    history_size: int
    max_ground_accel: float
    max_ground_velocity: float
    observation_noise: List[float]
    reward_weights: dict
    spine_config: dict
    spine_frequency: int


@gin.configurable
@dataclass
class PPOSettings:
    """PPO algorithm settings."""

    batch_size: int
    clip_range: float
    clip_range_vf: Optional[float]
    ent_coef: float
    gae_lambda: float
    learning_rate: float
    max_grad_norm: float
    n_epochs: int
    n_steps: int
    net_arch_pi: Tuple[int, int]
    net_arch_vf: Tuple[int, int]
    normalize_advantage: bool
    sde_sample_freq: int
    target_kl: Optional[float]
    use_sde: bool
    vf_coef: float


@gin.configurable
@dataclass
class TrainingSettings:
    """Training settings."""

    init_rand: dict
    max_episode_duration: float
    return_horizon: float
    total_timesteps: int
